"""TcEx Framework"""
# standard library
import inspect
import logging
import os
import platform
import re
import signal
import sys
import threading
from typing import Optional
from urllib.parse import quote

from .app_config_object import InstallJson
from .inputs import Inputs
from .logger import Logger
from .tokens import Tokens


class TcEx:
    """Provides basic functionality for all types of TxEx Apps.

    Args:
        config (dict, kwargs): A dictionary containing configuration items typically used by
            external Apps.
        config_file (str, kwargs): A filename containing JSON configuration items typically used
            by external Apps.
        logger (logging.Logger, kwargs): An pre-configured instance of logger to use instead of
            tcex logger.
    """

    def __init__(self, **kwargs):
        """Initialize Class Properties."""
        # catch interupt signals
        if threading.current_thread().name == 'MainThread':
            signal.signal(signal.SIGINT, self._signal_handler)
            if platform.system() != 'Windows':
                signal.signal(signal.SIGHUP, self._signal_handler)
            signal.signal(signal.SIGTERM, self._signal_handler)

        # Property defaults
        self._config: dict = kwargs.get('config') or {}
        self._default_args = None
        self._error_codes = None
        self._exit_code = 0
        self._indicator_associations_types_data = {}
        self._indicator_types = None
        self._indicator_types_data = None
        self._jobs = None
        self._logger = None
        self._playbook = None
        self._redis_client = None
        self._service = None
        self._session = None
        self._session_external = None
        self._stix_model = None
        self._utils = None
        self._ti = None
        self._token = None
        self.ij = InstallJson()

        # add custom logger if provided
        self._log: object = kwargs.get('logger')

        # init args (needs logger)
        self.inputs = Inputs(self, self._config, kwargs.get('config_file'))

    def _association_types(self):
        """Retrieve Custom Indicator Associations types from the ThreatConnect API."""
        # Dynamically create custom indicator class
        r: object = self.session.get('/v2/types/associationTypes')

        # check for bad status code and response that is not JSON
        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            self.log.warning('feature=tcex, event=association-types-download, status=failure')
            return

        # validate successful API results
        data: dict = r.json()
        if data.get('status') != 'Success':
            self.log.warning('feature=tcex, event=association-types-download, status=failure')
            return

        try:
            # Association Type Name is not a unique value at this time, but should be.
            for association in data.get('data', {}).get('associationType', []):
                self._indicator_associations_types_data[association.get('name')] = association
        except Exception as e:
            self.handle_error(200, [e])

    def _signal_handler(
        self, signal_interupt: int, frame: object  # pylint: disable=unused-argument
    ) -> None:
        """Handle singal interrupt."""
        call_file: str = os.path.basename(inspect.stack()[1][0].f_code.co_filename)
        call_module: str = inspect.stack()[1][0].f_globals['__name__'].lstrip('Functions.')
        call_line: int = inspect.stack()[1][0].f_lineno
        self.log.error(
            f'App interrupted - file: {call_file}, method: {call_module}, line: {call_line}.'
        )
        if signal_interupt in (2, 15):
            self.exit(1, 'The App received an interrupt signal and will now exit.')

    def advanced_request(self, session: object, timeout: Optional[int] = 600) -> object:
        """Return instance of AdvancedRequest.

        Args:
            session (object): An instance of requests.Session.
            timeout (int): The number of second before timing out the request.

        Returns:
            object: An instance of AdvancedRequest
        """
        from .app_feature import AdvancedRequest

        return AdvancedRequest(session, self, timeout)

    def aot_rpush(self, exit_code: int) -> None:
        """Push message to AOT action channel."""
        if self.default_args.tc_playbook_db_type == 'Redis':
            try:
                self.redis_client.rpush(self.default_args.tc_exit_channel, exit_code)
            except Exception as e:  # pragma: no cover
                self.exit(1, f'Exception during AOT exit push ({e}).')

    @property
    def args(self) -> object:
        """Argparser args Namespace."""
        return self.inputs.args()

    def batch(
        self,
        owner: str,
        action: Optional[str] = 'Create',
        attribute_write_type: Optional[str] = 'Replace',
        halt_on_error: Optional[bool] = False,
        playbook_triggers_enabled: Optional[bool] = False,
    ) -> object:
        """Return instance of Batch

        Args:
            tcex: An instance of TcEx object.
            owner: The ThreatConnect owner for Batch action.
            action: Action for the batch job ['Create', 'Delete'].
            attribute_write_type: Write type for TI attributes ['Append', 'Replace'].
            halt_on_error: If True any batch error will halt the batch job.
            playbook_triggers_enabled: Deprecated input, will not be used.

        Returns:
            object: An instance of the Batch Class.
        """
        from .batch import Batch

        return Batch(
            self, owner, action, attribute_write_type, halt_on_error, playbook_triggers_enabled
        )

    def cache(
        self,
        domain: str,
        data_type: str,
        ttl_seconds: Optional[int] = None,
        mapping: Optional[dict] = None,
    ) -> object:
        """Get instance of the Cache module.

        Args:
            domain: The domain can be either "system", "organization", or "local". When using
                "organization" the data store can be accessed by any Application in the entire org,
                while "local" access is restricted to the App writing the data. The "system" option
                should not be used in almost all cases.
            data_type: The data type descriptor (e.g., tc:whois:cache).
            ttl_seconds: The number of seconds the cache is valid.
            mapping: Advanced - The datastore mapping if required.

        Returns:
            object: An instance of the Cache Class.
        """
        from .datastore import Cache

        return Cache(self, domain, data_type, ttl_seconds, mapping)

    @property
    def case_management(self) -> object:
        """Include the Threat Intel Module.

        .. Note:: Threat Intell methods can be accessed using ``tcex.ti.<method>``.

        Returns:
            object: An instance of the CaseManagement Class.
        """
        from .case_management import CaseManagement

        return CaseManagement(self)

    @property
    def cm(self) -> object:
        """Include the Case Management Module."""
        return self.case_management

    def datastore(self, domain: str, data_type: str, mapping: Optional[dict] = None) -> object:
        """Get instance of the DataStore module.

        Args:
            domain: The domain can be either "system", "organization", or "local". When using
                "organization" the data store can be accessed by any Application in the entire org,
                while "local" access is restricted to the App writing the data. The "system" option
                should not be used in almost all cases.
            data_type: The data type descriptor (e.g., tc:whois:cache).
            mapping: ElasticSearch mappings data.

        Returns:
            object: An instance of the DataStore Class.
        """
        from .datastore import DataStore

        return DataStore(self, domain, data_type, mapping)

    @property
    def default_args(self) -> object:
        """Argparser args Namespace."""
        return self._default_args

    @property
    def error_codes(self) -> object:
        """Return TcEx error codes."""
        if self._error_codes is None:
            from .tcex_error_codes import TcExErrorCodes

            self._error_codes = TcExErrorCodes()
        return self._error_codes

    def exit(self, code: Optional[int] = None, msg: Optional[str] = None) -> None:
        """Application exit method with proper exit code

        The method will run the Python standard sys.exit() with the exit code
        previously defined via :py:meth:`~tcex.tcex.TcEx.exit_code` or provided
        during the call of this method.

        Args:
            code: The exit code value for the app.
            msg: A message to log and add to message tc output.
        """
        # add exit message to message.tc file and log
        if msg is not None:
            if code in [0, 3] or (code is None and self.exit_code in [0, 3]):
                self.log.info(msg)
            else:
                self.log.error(msg)
            self.message_tc(msg)

        if code is None:
            code = self.exit_code
        elif code in [0, 1, 3]:
            pass
        else:
            self.log.error('Invalid exit code')
            code = 1

        if self.default_args.tc_aot_enabled:
            # push exit message
            self.aot_rpush(code)

        # exit token renewal thread
        self.token.shutdown = True

        self.log.info(f'Exit Code: {code}')
        sys.exit(code)

    @property
    def exit_code(self) -> None:
        """Return the current exit code."""
        return self._exit_code

    @exit_code.setter
    def exit_code(self, code: int) -> None:
        """Set the App exit code.

        For TC Exchange Apps there are 3 supported exit codes.
        * 0 indicates a normal exit
        * 1 indicates a failure during execution
        * 3 indicates a partial failure

        Args:
            code (int): The exit code value for the app.
        """
        if code is not None and code in [0, 1, 3]:
            self._exit_code = code
        else:
            self.log.warning('Invalid exit code')

    @staticmethod
    def expand_indicators(indicator: str) -> list:
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator: An " : " delimited string.

        Returns:
            (list): a list of indicators split on " : ".
        """
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            # group 1 - lazy capture everything to first <space>:<space> or end of line
            iregx_pattern = r'^(.*?(?=\s\:\s|$))?'
            iregx_pattern += r'(?:\s\:\s)?'  # remove <space>:<space>
            # group 2 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead (optional <space>):<space> or end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
            iregx_pattern += r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
            # group 3 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=$))?$'
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())
        else:
            # handle all single valued indicator types (address, host, etc)
            indicator_list = [indicator]

        return indicator_list

    @property
    def victim_asset_types(self) -> list:
        """Return all defined ThreatConnect Asset types.

        Returns:
            (list): A list of ThreatConnect Asset types.
        """
        return [
            'EmailAddress',
            'SocialNetwork',
            'NetworkAccount',
            'WebSite',
            'Phone',
        ]

    @property
    def group_types(self) -> list:
        """Return all defined ThreatConnect Group types.

        Returns:
            (list): A list of ThreatConnect Group types.
        """
        return [
            'Adversary',
            'Campaign',
            'Document',
            'Email',
            'Event',
            'Incident',
            'Intrusion Set',
            'Signature',
            'Report',
            'Threat',
            'Task',
        ]

    @property
    def group_types_data(self) -> dict:
        """Return supported ThreatConnect Group types."""
        return {
            'Adversary': {'apiBranch': 'adversaries', 'apiEntity': 'adversary'},
            'Campaign': {'apiBranch': 'campaigns', 'apiEntity': 'campaign'},
            'Document': {'apiBranch': 'documents', 'apiEntity': 'document'},
            'Email': {'apiBranch': 'emails', 'apiEntity': 'email'},
            'Event': {'apiBranch': 'events', 'apiEntity': 'event'},
            'Incident': {'apiBranch': 'incidents', 'apiEntity': 'incident'},
            'Intrusion Set': {'apiBranch': 'intrusionSets', 'apiEntity': 'intrusionSet'},
            'Report': {'apiBranch': 'reports', 'apiEntity': 'report'},
            'Signature': {'apiBranch': 'signatures', 'apiEntity': 'signature'},
            'Threat': {'apiBranch': 'threats', 'apiEntity': 'threat'},
            'Task': {'apiBranch': 'tasks', 'apiEntity': 'task'},
        }

    def get_type_from_api_entity(self, api_entity: dict) -> Optional[str]:
        """Return the object type as a string given a api entity.

        Args:
            api_entity: A TCEntity object.

        Returns:
            str, None: The type value or None.

        """
        merged = self.group_types_data.copy()
        merged.update(self.indicator_types_data)
        for key, value in merged.items():
            if value.get('apiEntity') == api_entity:
                return key
        return None

    def handle_error(
        self, code: int, message_values: Optional[list] = None, raise_error: Optional[bool] = True
    ) -> None:
        """Raise RuntimeError

        Args:
            code: The error code from API or SDK.
            message: The error message from API or SDK.
            raise_error: Raise a Runtime error. Defaults to True.

        Raises:
            RuntimeError: Raised a defined error.
        """
        try:
            if message_values is None:
                message_values = []
            message = self.error_codes.message(code).format(*message_values)
            self.log.error(f'Error code: {code}, {message}')
        except AttributeError:
            self.log.error(f'Incorrect error code provided ({code}).')
            raise RuntimeError(100, 'Generic Failure, see logs for more details.')
        except IndexError:
            self.log.error(
                f'Incorrect message values provided for error code {code} ({message_values}).'
            )
            raise RuntimeError(100, 'Generic Failure, see logs for more details.')
        if raise_error:
            raise RuntimeError(code, message)

    @property
    def indicator_associations_types_data(self) -> dict:
        """Return ThreatConnect associations type data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dict): A dictionary of ThreatConnect associations types.
        """
        if not self._indicator_associations_types_data:
            self._association_types()  # load custom indicator associations
        return self._indicator_associations_types_data

    @property
    def indicator_types(self) -> list:
        """Return ThreatConnect Indicator types.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (list): A list of ThreatConnect Indicator types.
        """
        if not self._indicator_types:
            self._indicator_types = self.indicator_types_data.keys()
        return self._indicator_types

    @property
    def indicator_types_data(self) -> dict:
        """Return ThreatConnect indicator types data.

        Retrieve the data from the API if it hasn't already been retrieved.

        Returns:
            (dict): A dictionary of ThreatConnect Indicator data.
        """
        if not self._indicator_types_data:
            self._indicator_types_data = {}

            # retrieve data from API
            r = self.session.get('/v2/types/indicatorTypes')
            # TODO: use handle error instead
            if not r.ok:
                raise RuntimeError('Could not retrieve indicator types from ThreatConnect API.')

            for itd in r.json().get('data', {}).get('indicatorType'):
                self._indicator_types_data[itd.get('name')] = itd
        return self._indicator_types_data

    @property
    def log(self) -> object:
        """Return a valid logger."""
        if self._log is None:
            self._log = self.logger.log
        return self._log

    @log.setter
    def log(self, log: object) -> None:
        """Return a valid logger."""
        if isinstance(log, logging.Logger):
            self._log = log

    @property
    def logger(self) -> object:
        """Return logger."""
        if self._logger is None:
            logger_name = self._config.get('tc_logger_name', 'tcex')
            self._logger = Logger(self, logger_name)
            self._logger.add_cache_handler('cache')
        return self._logger

    def metric(
        self,
        name: str,
        description: str,
        data_type: str,
        interval: str,
        keyed: Optional[bool] = False,
    ) -> object:
        """Get instance of the Metrics module.

        Args:
            name: The name for the metric.
            description: The description of the metric.
            data_type: The type of metric: Sum, Count, Min, Max, First, Last, and Average.
            interval: The metric interval: Hourly, Daily, Weekly, Monthly, and Yearly.
            keyed: Indicates whether the data will have a keyed value.

        Returns:
            (object): An instance of the Metrics Class.
        """
        from .metrics import Metrics

        return Metrics(self, name, description, data_type, interval, keyed)

    def message_tc(self, message: str, max_length: Optional[int] = 255) -> None:
        """Write data to message_tc file in TcEX specified directory.

        This method is used to set and exit message in the ThreatConnect Platform.
        ThreatConnect only supports files of max_message_length.  Any data exceeding
        this limit will be truncated. The last <max_length> characters will be preserved.

        Args:
            message: The message to add to message_tc file
            max_length: The maximum length of an exit message. Defaults to 255.
        """
        if not isinstance(message, str):
            message = str(message)

        if os.access(self.default_args.tc_out_path, os.W_OK):
            message_file = os.path.join(self.default_args.tc_out_path, 'message.tc')
        else:
            message_file = 'message.tc'

        if os.path.isfile(message_file):
            with open(message_file) as mh:
                message = mh.read() + message

        if not message.endswith('\n'):
            message += '\n'
        with open(message_file, 'w') as mh:
            # write last <max_length> characters to file
            mh.write(message[-max_length:])

    def notification(self) -> object:
        """Get instance of the Notification module.

        Returns:
            (object): An instance of the Notification Class.
        """
        from .notifications import Notifications

        return Notifications(self)

    @property
    def parser(self) -> object:
        """Instance tcex args parser."""
        return self.inputs.parser

    def pb(self, context: str, output_variables: list) -> object:
        """Return a new instance of playbook module.

        Args:
            context: The Redis context for Playbook or Service Apps.
            output_variables: A list of requested PB/Service output variables.

        Returns:
            tcex.playbook.Playbooks: An instance of Playbooks
        """
        from .playbooks import Playbooks

        return Playbooks(self, context, output_variables)

    @property
    def playbook(self) -> object:
        """Return an instance of Playbooks module.

        This property defaults context and outputvariables to arg values.

        .. Note:: Playbook methods can be accessed using ``tcex.playbook.<method>``.

        Returns:
            tcex.playbook.Playbooks: An instance of Playbooks
        """
        if self._playbook is None:
            # handle outputs coming in as a csv string and list
            outputs: list = self.default_args.tc_playbook_out_variables or []
            if isinstance(outputs, str):
                outputs = outputs.split(',')
            self._playbook: object = self.pb(self.default_args.tc_playbook_db_context, outputs)
        return self._playbook

    @property
    def proxies(self) -> dict:
        """Format the proxy configuration for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {"http": "http://user:pass@10.10.1.10:3128/"}

        Returns:
           (dict): Dictionary of proxy settings
        """
        proxies = {}
        if (
            self.default_args.tc_proxy_host is not None
            and self.default_args.tc_proxy_port is not None
        ):

            if (
                self.default_args.tc_proxy_username is not None
                and self.default_args.tc_proxy_password is not None
            ):
                tc_proxy_username = quote(self.default_args.tc_proxy_username, safe='~')
                tc_proxy_password = quote(self.default_args.tc_proxy_password, safe='~')

                # proxy url with auth
                proxy_url = (
                    f'{tc_proxy_username}:{tc_proxy_password}'
                    f'@{self.default_args.tc_proxy_host}:{self.default_args.tc_proxy_port}'
                )
            else:
                # proxy url without auth
                proxy_url = f'{self.default_args.tc_proxy_host}:{self.default_args.tc_proxy_port}'
            proxies = {'http': f'http://{proxy_url}', 'https': f'https://{proxy_url}'}
        return proxies

    @property
    def rargs(self) -> object:
        """Return argparser args Namespace with Playbook args automatically resolved."""
        return self.inputs.resolved_args()

    @staticmethod
    def rc(host, port, db=0, blocking=False, **kwargs) -> object:
        """Return a *new* instance of Redis client.

        For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

        Args:
            host (str, optional): The REDIS host. Defaults to localhost.
            port (int, optional): The REDIS port. Defaults to 6379.
            db (int, optional): The REDIS db. Defaults to 0.
            blocking_pool (bool): Use BlockingConnectionPool instead of ConnectionPool.
            errors (str, kwargs): The REDIS errors policy (e.g. strict).
            max_connections (int, kwargs): The maximum number of connections to REDIS.
            password (str, kwargs): The REDIS password.
            socket_timeout (int, kwargs): The REDIS socket timeout.
            timeout (int, kwargs): The REDIS Blocking Connection Pool timeout value.

        Returns:
            Redis.client: An instance of redis client.
        """
        from .key_value_store import RedisClient

        return RedisClient(host=host, port=port, db=db, blocking=blocking, **kwargs).client

    @property
    def redis_client(self) -> object:
        """Return redis client instance configure for Playbook/Service Apps."""
        if self._redis_client is None:
            from .key_value_store import RedisClient

            self._redis_client = RedisClient(
                host=self.default_args.tc_playbook_db_path,
                port=self.default_args.tc_playbook_db_port,
                db=0,
            ).client

        return self._redis_client

    def results_tc(self, key: str, value: str) -> None:
        """Write data to results_tc file in TcEX specified directory.

        The TcEx platform support persistent values between executions of the App.  This
        method will store the values for TC to read and put into the Database.

        Args:
            key: The data key to be stored.
            value: The data value to be stored.
        """
        if os.access(self.default_args.tc_out_path, os.W_OK):
            results_file = f'{self.default_args.tc_out_path}/results.tc'
        else:
            results_file = 'results.tc'

        new = True
        open(results_file, 'a').close()  # ensure file exists
        with open(results_file, 'r+') as fh:
            results = ''
            for line in fh.read().strip().split('\n'):
                if not line:
                    continue
                try:
                    k, v = line.split(' = ')
                except ValueError:
                    # handle null/empty value (e.g., "name =")
                    k, v = line.split(' =')
                if k == key:
                    v = value
                    new = False
                if v is not None:
                    results += f'{k} = {v}\n'
            if new and value is not None:  # indicates the key/value pair didn't already exist
                results += f'{key} = {value}\n'
            fh.seek(0)
            fh.write(results)
            fh.truncate()

    @staticmethod
    def safe_indicator(indicator: str) -> str:
        """Format indicator value for safe HTTP request.

        Args:
            indicator: Indicator to URL Encode

        Returns:
            (str): The urlencoded string
        """
        if indicator is not None:
            indicator = quote(indicator, safe='~')
        return indicator

    @staticmethod
    def safe_rt(resource_type: str, lower: Optional[bool] = False) -> str:
        """Format the Resource Type.

        Takes Custom Indicator types with a space character and return a *safe* string.

        (e.g. *User Agent* is converted to User_Agent or user_agent.)

        Args:
           resource_type: The resource type to format.
           lower: Return type in all lower case

        Returns:
            (str): The formatted resource type.
        """
        if resource_type is not None:
            resource_type = resource_type.replace(' ', '_')
            if lower:
                resource_type = resource_type.lower()
        return resource_type

    @staticmethod
    def safe_group_name(
        group_name: str, group_max_length: Optional[int] = 100, ellipsis: Optional[bool] = True
    ):
        """Truncate group name to match limit breaking on space and optionally add an ellipsis.

        .. note:: Currently the ThreatConnect group name limit is 100 characters.

        Args:
           group_name: The raw group name to be truncated.
           group_max_length: The max length of the group name.
           ellipsis: If true the truncated name will have '...' appended.

        Returns:
            (str): The truncated group name with optional ellipsis.
        """
        ellipsis_value = ''
        if ellipsis:
            ellipsis_value = ' ...'

        if group_name is not None and len(group_name) > group_max_length:
            # split name by spaces and reset group_name
            group_name_array = group_name.split(' ')
            group_name = ''
            for word in group_name_array:
                word = f'{word}'
                if (len(group_name) + len(word) + len(ellipsis_value)) >= group_max_length:
                    group_name = f'{group_name}{ellipsis_value}'
                    group_name = group_name.lstrip(' ')
                    break
                group_name += f' {word}'
        return group_name

    @staticmethod
    def safe_tag(tag: str) -> str:
        """Encode and truncate tag to match limit (128 characters) of ThreatConnect API.

        Args:
           tag: The tag to be truncated

        Returns:
            (str): The truncated and quoted tag.
        """
        if tag is not None:
            tag = quote(tag[:128], safe='~')
        return tag

    @staticmethod
    def safe_url(url: str) -> str:
        """Encode value for safe HTTP request.

        Args:
            url (str): The string to URL Encode.

        Returns:
            (str): The urlencoded string.
        """
        if url is not None:
            url: str = quote(url, safe='~')
        return url

    @property
    def service(self) -> object:
        """Include the Service Module.

        .. Note:: Service methods can be accessed using ``tcex.service.<method>``.
        """
        if self._service is None:
            if self.ij.runtime_level.lower() == 'apiservice':
                from .services import ApiService as Service
            elif self.ij.runtime_level.lower() == 'triggerservice':
                from .services import CommonServiceTrigger as Service
            elif self.ij.runtime_level.lower() == 'webhooktriggerservice':
                from .services import WebhookTriggerService as Service
            else:
                self.exit(1, 'Could not determine the service type.')

            self._service = Service(self)
        return self._service

    @property
    def session(self) -> object:
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        if self._session is None:
            from .sessions import TcSession

            self._session = TcSession(
                logger=self.log,
                api_access_id=self.default_args.api_access_id,
                api_secret_key=self.default_args.api_secret_key,
                base_url=self.default_args.tc_api_path,
            )

            # set verify
            self._session.verify = self.default_args.tc_verify

            # set token
            self._session.token = self.token

            # update User-Agent
            self._session.headers.update(
                {'User-Agent': f'TcEx: {__import__(__name__).__version__}'}
            )

            # add proxy support if requested
            if self.default_args.tc_proxy_tc:
                self._session.proxies = self.proxies
                self.log.info(
                    f'Using proxy host {self.args.tc_proxy_host}:'
                    f'{self.args.tc_proxy_port} for ThreatConnect session.'
                )

            # enable curl logging if tc_log_curl param is set.
            if self.default_args.tc_log_curl:
                self._session.log_curl = True
        return self._session

    @property
    def session_external(self) -> object:
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        if self._session_external is None:
            from .sessions import ExternalSession

            self._session_external = ExternalSession(logger=self.log)

            # add User-Agent to headers
            self._session_external.headers.update(
                {'User-Agent': f'TcEx App: {self.ij.display_name} - {self.ij.program_version}'}
            )

            # add proxy support if requested
            if self.default_args.tc_proxy_external:
                self._session_external.proxies = self.proxies
                self.log.info(
                    f'Using proxy host {self.args.tc_proxy_host}:'
                    f'{self.args.tc_proxy_port} for external session.'
                )

            if self.default_args.tc_log_curl:
                self._session_external.log_curl = True
        return self._session_external

    @property
    def stix_model(self) -> object:
        """Include the Threat Intel Module.

        .. Note:: Threat Intell methods can be accessed using ``tcex.ti.<method>``.
        """
        if self._stix_model is None:
            from .stix import StixModel

            self._stix_model = StixModel(self.logger)
        return self._stix_model

    @property
    def ti(self) -> object:
        """Include the Threat Intel Module.

        .. Note:: Threat Intell methods can be accessed using ``tcex.ti.<method>``.
        """
        if self._ti is None:
            from .threat_intelligence import ThreatIntelligence

            self._ti = ThreatIntelligence(self)
        return self._ti

    @property
    def token(self) -> object:
        """Return token object."""
        if self._token is None:
            sleep_interval = int(os.getenv('TC_TOKEN_SLEEP_INTERVAL', '30'))
            self._token = Tokens(
                self.default_args.tc_api_path, sleep_interval, self.default_args.tc_verify, self.log
            )
        return self._token

    @property
    def utils(self) -> object:
        """Include the Utils module.

        .. Note:: Utils methods can be accessed using ``tcex.utils.<method>``.
        """
        if self._utils is None:
            from .utils import Utils

            self._utils = Utils(temp_path=self.default_args.tc_temp_path)
        return self._utils



"""Defines Visitor and VisitorPattern for STIX or TC Objects.

Visitor and Visitor producer should be used when parsing isn't a simple mapping from the source
format to the target format, and instead needs to modify the result of other parsers. For example,
A STIX Relationship object defines and association between two other objects.
"""
# standard library
from typing import Iterable, List, Union


class Visitor:
    """Visitors receive the final generator of parsed data from StixModel and can modify members.

    StixModel -[]-[]-[]-[]--> |VISITOR A| -[X]-[]-[X]--> |VISITOR B| -[XY]-[Y]-[X]--> Caller
    """

    def visit(self, data: Iterable[dict]) -> Iterable[dict]:  # pylint: disable=no-self-use
        """Receive the final generator of parsed data from StixModel, and optionally modify it.

        Args:
            data:  All of the parsed data.

        Returns:
            Generator of data, with any modifications.
        """
        yield from data


class VisitorProducer:
    """Creates and returns a visitor."""

    def consume(
        self, stix_data: Union[List[dict], dict]
    ) -> Iterable[Visitor]:  # pylint: disable=no-self-use
        """Parse STIX data and return a Visitor.

        Args:
            stix_data: One or more STIX objects to parse.

        Returns:
            A visitor that will apply the data parsed from the STIX objects to final parsed objects
            generator.
        """

    def produce(
        self, tc_data: Union[List[dict], dict]
    ) -> Iterable[Visitor]:  # pylint: disable=no-self-use
        """Parse ThreatConnect data and return a Visitor.

        Args:
            tc_data: One or more ThreatConnect objects to parse.

        Returns:
            A visitor that will apply the data parsed from the ThreatConnect object(s) to final
            parsed objects generator.
        """



"""TcEx Framework Env Store Module"""
# standard library
import logging
import os
import sys

try:
    # third-party
    import hvac
except ImportError:
    # hvac on required for local dev/testing
    pass


class EnvStore:
    """TcEx Key Value API Module.

    Args:
        logger (logging.Logger, optional): A instance of Logger. Defaults to None.
    """

    def __init__(self, logger=None):
        """Initialize the Class properties."""
        self.log = logger or logging.getLogger('layout_json')

        # properties
        self._vault_client = None
        self.cache = {}
        self.vault_token = os.getenv('VAULT_TOKEN')
        self.vault_addr = os.getenv('VAULT_ADDR') or os.getenv('VAULT_URL')

    @staticmethod
    def _convert_env(env_variable):
        """Convert an a vault path to env variable, removing first 2 parts of path

        Vault path need to be updated to be looked up as env variables.

        /ninja/int/cisco/umbrella/token -> CISCO_UMBRELLA_TOKEN

        """
        if '/' in env_variable:
            paths = env_variable.lstrip('/').split('/')
            # remove first to parts of path for env variable and leading '/'
            env_variable = '/'.join(paths[2:])
        return env_variable.replace('/', '_').replace(' ', '_').upper()

    def getenv(self, env_variable, env_type='env', default=None):
        """Return the value for the provide environment variable.

        Args:
            env_variable (str): The env variable name or env store path.
            env_type (str, optional): The type of environment variable to look up. Defaults
                to 'env'.
            default (str): The default value if no value is found.

        Returns:
            str|None: The value from OS env or env store.
        """
        cache_key = f'{env_type}-{env_variable}'
        cache_value = self.cache.get(cache_key)
        if cache_value is not None:
            return cache_value

        env_var_updated = self._convert_env(env_variable)
        value = default

        if env_type in ['env', 'envs', 'local'] and os.getenv(env_var_updated):
            # return value from OS environ
            value = os.getenv(env_var_updated, default)
        elif env_type in ['env', 'envs', 'remote'] and 'hvac' in sys.modules:
            # return value from Vault
            value = self.read_from_vault(env_variable, default)

        # provide an error so dev/qa engineer knows that
        # an env var they provide could not be found
        if value is None:
            raise RuntimeError(
                f'Could not resolve env variable {env_variable} ({env_var_updated}).'
            )

        self.cache[cache_key] = value
        return value

    def read_from_vault(self, full_path, default=None):
        """Read data from Vault for the provided path.

        Args:
            full_path (string): The path to the vault data including the key
                (e.g. myData/mySecret/myKey).
            default (str): The default value if no value is found.

        Returns:
            str|None: The vault association with the provided path and key.
        """
        paths = full_path.lstrip('/').split('/')

        # the key stored in data object at the provided path
        # (e.g., "/myData/myResource/token" -> "token")
        key = paths[-1].strip('/')

        # the path with the key and mount point removed
        # (e.g., "/myData/myResource/token/" -> "myResource")
        path = '/'.join(paths[1:-1])

        # the mount point from the path
        # (e.g., "/myData/myResource/token/" -> "myData")
        mount_point = paths[0]

        data = {}
        try:
            data = self.vault_client.secrets.kv.read_secret_version(
                path=path, mount_point=mount_point
            )
        except hvac.exceptions.InvalidPath:
            if hasattr(self.log, 'data'):
                self.log.data('setup', 'env store', f'Path not found: {path}')
        except hvac.exceptions.VaultError as e:
            if hasattr(self.log, 'data'):
                self.log.data('setup', 'env store', f'Error reading path: "{path}" ({e})', 'error')
        except Exception:
            self.log.data(
                'setup',
                'env store',
                'Generic failure. Please ensure Environment Store '
                'connectivity and env are configured properly.',
                'error',
            )

        if hasattr(self.log, 'data'):
            self.log.data('setup', 'vault read', full_path, 'trace')

        return data.get('data', {}).get('data', {}).get(key) or default

    @property
    def vault_client(self):
        """Return configured vault client."""
        if self._vault_client is None and 'hvac' in sys.modules:
            if self.vault_addr is not None and self.vault_token is not None:
                self._vault_client = hvac.Client(url=self.vault_addr, token=self.vault_token)
        return self._vault_client



"""ThreatConnect Workflow Event"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class WorkflowEvents(CommonCaseManagementCollection):
    """Workflow Events Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Artifact. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Workflow Event objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.WORKFLOW_EVENTS,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

    def __iter__(self):
        """Iterate on Workflow Events."""
        return self.iterate(initial_response=self.initial_response)

    def entity_map(self, entity):
        """Map a dict to a Workflow Event.

        Args:
            entity (dict): The Workflow Event data.

        Returns:
            CaseManagement.WorkflowEvent: An Workflow Event Object
        """
        return WorkflowEvent(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterWorkflowEvent Object."""
        return FilterWorkflowEvents(ApiEndpoints.WORKFLOW_EVENTS, self.tcex, self.tql)


class WorkflowEvent(CommonCaseManagement):
    """WorkflowEvent object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Workflow Event.
        case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Workflow Event.
        date_added (str, kwargs): [Read-Only] The **Date Added** for the Workflow Event.
        deleted (bool, kwargs): [Read-Only] The **Deleted** flag for the Workflow Event.
        deleted_reason (str, kwargs): the reason for deleting the event (required input for DELETE
            operation only)
        event_date (str, kwargs): the time that the Event is logged
        link (str, kwargs): [Read-Only] The **Link** for the Workflow Event.
        link_text (str, kwargs): [Read-Only] The **Link Text** for the Workflow Event.
        notes (Note, kwargs): a list of Notes corresponding to the Event
        parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Workflow Event.
        summary (str, kwargs): [Required] The **Summary** for the Workflow Event.
        system_generated (bool, kwargs): [Read-Only] The **System Generated** flag for the Workflow
            Event.
        user (User, kwargs): [Read-Only] The **User** for the Workflow Event.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.WORKFLOW_EVENTS, kwargs)
        self.workflow_event_filter = [
            {
                'keyword': 'workflowEventId',
                'operator': TQL.Operator.EQ,
                'value': self.id,
                'type': TQL.Type.INTEGER,
            }
        ]
        self._case_id = kwargs.get('case_id', None) or kwargs.get('parent_case', {}).get('id', None)
        self._case_xid = kwargs.get('case_xid', None) or kwargs.get('parent_case', {}).get(
            'xid', None
        )
        self._date_added = kwargs.get('date_added', None)
        self._deleted = kwargs.get('deleted', None)
        self._deleted_reason = kwargs.get('deleted_reason', None)
        self._event_date = kwargs.get('event_date', None)
        self._link = kwargs.get('link', None)
        self._link_text = kwargs.get('link_text', None)
        self._notes = kwargs.get('notes', None)
        self._parent_case = kwargs.get('parent_case', None)
        self._summary = kwargs.get('summary', None)
        self._system_generated = kwargs.get('system_generated', None)
        self._user = kwargs.get('user', None)

    def add_note(self, **kwargs):
        """Add a note to the workflow event."""
        self.notes.add_note(self.tcex.cm.note(**kwargs))

    @property
    def as_entity(self):
        """Return the entity representation of the Workflow Event."""
        return {'type': 'Workflow Event', 'value': self.summary, 'id': self.id}

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_workflow_event = WorkflowEvent(self.tcex, **entity)
        self.__dict__.update(new_workflow_event.__dict__)

    @property
    def case_id(self):
        """Return the **Case ID** for the Workflow Event."""
        return self._case_id

    @case_id.setter
    def case_id(self, case_id):
        """Set the **Case ID** for the Workflow Event."""
        self._case_id = case_id

    @property
    def case_xid(self):
        """Return the **Case XID** for the Workflow Event."""
        return self._case_xid

    @case_xid.setter
    def case_xid(self, case_xid):
        """Set the **Case XID** for the Workflow Event."""
        self._case_xid = case_xid

    @property
    def date_added(self):
        """Return the **Date Added** for the Workflow Event."""
        return self._date_added

    @property
    def deleted(self):
        """Return the **Deleted** flag for the Workflow Event."""
        return self._deleted

    @property
    def deleted_reason(self):
        """Return the **Deleted Reason** for the Workflow Event."""
        return self._deleted_reason

    # TODO: @mj - confirm the status of this field
    # @deleted_reason.setter
    # def deleted_reason(self, deleted_reason):
    #     """Set the **Deleted Reason** for the Workflow Event."""
    #     self._deleted_reason = deleted_reason

    @property
    def event_date(self):
        """Return the **Event Date** for the Workflow Event."""
        return self._event_date

    @event_date.setter
    def event_date(self, event_date):
        """Set the **Event Date** for the Workflow Event."""
        self._event_date = event_date

    @property
    def id(self):
        """Return the **ID** for the Workflow Event."""
        return self._id

    @id.setter
    def id(self, id):  # pylint: disable=redefined-builtin
        """Set the **ID** for the Workflow Event."""
        self._id = id

    @property
    def link(self):
        """Return the **Link** for the Workflow Event."""
        return self._link

    @property
    def link_text(self):
        """Return the **Link Text** for the Workflow Event."""
        return self._link_text

    @property
    def notes(self):
        """Return the **Notes** for the Task"""
        if self._notes is None or isinstance(self._notes, dict):
            notes = self._notes or {}
            self._notes = self.tcex.cm.notes(
                initial_response=notes, tql_filters=self.workflow_event_filter
            )
        return self._notes

    @notes.setter
    def notes(self, notes):
        """Set the **Notes** for the Task"""
        if isinstance(notes, dict):
            self._notes = self.tcex.cm.notes(
                initial_response=notes, tql_filters=self.workflow_event_filter
            )
        self._notes = notes

    @property
    def parent_case(self):
        """Return the **Parent Case** for the Task."""
        if self._parent_case:
            return self.tcex.cm.case(**self._parent_case)
        return self._parent_case

    @property
    def summary(self):
        """Return the "Summary" for the Workflow Event."""
        return self._summary

    @summary.setter
    def summary(self, summary):
        """Set the "Summary" for the Workflow Event."""
        self._summary = summary

    @property
    def system_generated(self):
        """Return the **System Generated** flag for the Workflow Event."""
        return self._system_generated

    @property
    def user(self):
        """Return the **User** for the Workflow Event."""
        if self._user:
            return self.tcex.cm.user(**self._user)
        return self._user


class FilterWorkflowEvents(Filter):
    """Filter Object for WorkflowEvents"""

    def case_id(self, operator, case_id):
        """Filter Workflow Events based on **caseId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_id (int): The ID of the case this event is associated with.
        """
        self._tql.add_filter('caseId', operator, case_id, TQL.Type.INTEGER)

    def date_added(self, operator, date_added):
        """Filter Workflow Events based on **dateAdded** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            date_added (str): The date the event was added.
        """
        self._tql.add_filter('dateAdded', operator, date_added, TQL.Type.STRING)

    def deleted(self, operator, deleted):
        """Filter Workflow Events based on **deleted** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            deleted (bool): The deletion status of the event.
        """
        self._tql.add_filter('deleted', operator, deleted, TQL.Type.BOOLEAN)

    # TODO: @mj - confirm the status of these fields
    # Response: https://threatconnect.slack.com/archives/GS2NQL5SP/p1579282668019800
    def deleted_reason(self, operator, deleted_reason):  # pragma: no cover
        """Filter Workflow Events based on **deletedReason** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            deleted_reason (str): The reason the event was deleted.
        """
        self._tql.add_filter('deletedReason', operator, deleted_reason, TQL.Type.STRING)

    def event_date(self, operator, event_date):
        """Filter Workflow Events based on **eventDate** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            event_date (str): The date the event occurred.
        """
        self._tql.add_filter('eventDate', operator, event_date, TQL.Type.STRING)

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Workflow Events based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the event.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def link(self, operator, link):
        """Filter objects based on "link" field.

        Args:
            operator (enum): The enum for the required operator.
            link (str): The filter value.
        """
        self._tql.add_filter('link', operator, link)

    def summary(self, operator, summary):
        """Filter Workflow Events based on **summary** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            summary (str): Text of the event.
        """
        self._tql.add_filter('summary', operator, summary, TQL.Type.STRING)

    def system_generated(self, operator, system_generated):
        """Filter Workflow Events based on **systemGenerated** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            system_generated (bool): Flag determining if this event was created automatically by
                the system.
        """
        self._tql.add_filter('systemGenerated', operator, system_generated, TQL.Type.BOOLEAN)

    def user_name(self, operator, user_name):
        """Filter Workflow Events based on **userName** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            user_name (str): The username associated with the event.
        """
        self._tql.add_filter('userName', operator, user_name, TQL.Type.STRING)



"""ThreatConnect TI Security Label"""

from .mappings import Mappings

# import local modules for dynamic reference
module = __import__(__name__)


class SecurityLabel(Mappings):
    """Unique API calls for SecurityLabel API Endpoints"""

    def __init__(self, tcex, name, **kwargs):
        """

        Args:
            tcex:
            name:
            **kwargs:
        """
        super().__init__(
            tcex,
            main_type='SecurityLabel',
            api_type='securitylabels',
            sub_type=None,
            api_entity='securityLabel',
            api_branch=None,
            owner=kwargs.pop('owner', None),
        )
        self._data['type'] = 'securityLabels'
        self._data['sub_type'] = None
        self._data['name'] = name

        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    @property
    def as_entity(self):
        """Return the object as an entity."""
        return {}

    @staticmethod
    def is_security_label():
        """Indicates that this is a security label object

        Returns:

        """
        return True

    def can_create(self):
        """If the name has been provided returns that the SecurityLabel can be created, otherwise
        returns that the SecurityLabel cannot be created.

        Returns:

        """
        if self._data.get('name'):
            return True
        return False

    def add_key_value(self, key, value):
        """Converts the value and adds it as a data field.

          Args:
              key:
              value:
          """
        self._data[key] = value

    def name(self, name):
        """Updates the security labels name.

        Args:
            name:
        """
        self._data['name'] = name
        data = {'name': name}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def color(self, color):
        """Updates the security labels color.

        Args:
            color:

        """
        self._data['color'] = color
        data = {'color': color}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def description(self, description):
        """Updates the security labels description.

        Args:
            description:
        """
        self._data['description'] = description
        data = {'description': description}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )

    def date_added(self, date_added):
        """Updates the security labels date_added

        Args:
            date_added: Converted to %Y-%m-%dT%H:%M:%SZ date format
        """
        date_added = self._utils.datetime.format_datetime(
            date_added, date_format='%Y-%m-%dT%H:%M:%SZ'
        )

        self._data['dateAdded'] = date_added
        data = {'dateAdded': date_added}
        return self._tc_requests.update(
            self.api_type, self.api_branch, self.unique_id, data, owner=self.owner
        )



"""ThreatConnect TI Victim"""

from .mappings import Mappings

# import local modules for dynamic reference
module = __import__(__name__)


class Victim(Mappings):
    """Unique API calls for Victim API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            owner (str, kwargs): The owner for this Victim. Default to default Org when not provided
            name (str, kwargs): [Required for Create] The name for this Victim.
        """
        super().__init__(tcex, 'Victim', 'victims', None, 'victim', None, kwargs.pop('owner', None))
        self.name = None
        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    def _set_unique_id(self, json_response):
        """Set the unique id of the Group."""
        self.unique_id = json_response.get('id', '')

    def add_asset(self, asset_type, body):
        """Add an asset to the Victim

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            asset_type: (str) Either email, network, phone, social, or web.
            body: (dict) the body of the asset being added.

        Return:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if body is None:
            body = {}

        if asset_type is None:
            self._tcex.handle_error(
                925, ['asset_type', 'update_asset', 'asset_type', 'asset_type', asset_type]
            )

        return self.tc_requests.victim_add_asset(self.unique_id, asset_type, body)

    def add_email_asset(self, address, address_type):
        """Add a email asset to the Victim

        Args:
            address: (str) The asset address.
            address_type: (str) The asset address_type

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {'address': address, 'addressType': address_type}
        return self.add_asset('EMAIL', asset_data)

    def add_key_value(self, key, value):
        """Add the key-value to the Victim. """
        key = self._metadata_map.get(key, key)
        if key in ['unique_id', 'id']:
            self._unique_id = str(value)
        else:
            self._data[key] = value

    @property
    def _metadata_map(self):
        """Return metadata map for Group objects."""
        return {'work_location': 'workLocation'}

    def add_network_asset(self, account, network):
        """Add a network asset to the Victim

        Args:
            account: (str) The asset account.
            network: (str) The asset network

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {'account': account, 'network': network}
        return self.add_asset('NETWORK', asset_data)

    def add_phone_asset(self, phone_type):
        """Add a phone asset to the Victim

        Args:
            phone_type: (str) The asset phone type.

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {'phoneType': phone_type}
        return self.add_asset('PHONE', asset_data)

    def add_social_asset(self, account, network):
        """Add a social asset to the Victim

        Args:
            account: (str) The asset account.
            network: (str) The asset network

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {'account': account, 'network': network}
        return self.add_asset('SOCIAL', asset_data)

    def add_web_asset(self, web_site):
        """Add a web asset to the Victim

        Args:
            web_site: (str) The asset account.

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {'webSite': web_site}
        return self.add_asset('WEB', asset_data)

    @property
    def as_entity(self):
        """Return the entity representation of the Victim."""
        return {
            'type': 'Victim',
            'value': self.name,
            'id': int(self.unique_id) if self.unique_id else None,
        }

    def assets(self, asset_type=None):
        """
        Gets the assets of a Victim

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            asset_type: (str) The type of asset to be retrieved. Defaults to all of them.

        Yield:
            Json: The asset being retrieved.

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.victim_assets(
            self.api_type, self.api_branch, self.unique_id, asset_type
        )

    def can_create(self):
        """Return True if victim can be create."""
        return self.data.get('name') is not None

    def delete_asset(self, asset_id, asset_type):
        """Delete an asset of the Victim

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            asset_id: (int) the id of the asset being deleted.
            asset_type: (str) Either email, network, phone, social, or web.

        Return:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if asset_type is None:
            self._tcex.handle_error(
                925, ['asset_type', 'update_asset', 'asset_type', 'asset_type', asset_type]
            )

        return self.tc_requests.victim_delete_asset(self.unique_id, asset_type, asset_id)

    def delete_email_asset(self, asset_id):
        """Delete an email asset of the Victim

        Args:
            asset_id: (int) the id of the asset being deleted.

        Return:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'EMAIL')

    def delete_network_asset(self, asset_id):
        """Delete an network asset of the Victim

        Args:
            asset_id: (int) the id of the asset being deleted.

        Return:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'NETWORK')

    def delete_phone_asset(self, asset_id):
        """Delete an phone asset of the Victim

        Args:
            asset_id: (int) the id of the asset being deleted.

        Return:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'PHONE')

    def delete_social_asset(self, asset_id):
        """Delete an social asset of the Victim

        Args:
            asset_id: (int) the id of the asset being deleted.

        Return:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'SOCIAL')

    def delete_web_asset(self, asset_id):
        """Delete an web asset of the Victim

        Args:
            asset_id: (int) the id of the asset being deleted.

        Return:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'WEB')

    def email_assets(self):
        """
        Gets the email assets of a Victim

        Yield:
            Json: The asset being retrieved.
        """
        return self.assets(asset_type='EMAIL')

    def get_asset(self, asset_id, asset_type):
        """
        Gets the assets of a Victim

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            asset_id: (int) the id of the asset being deleted.
            asset_type: (str) The type of asset to be retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        if asset_type is None:
            self._tcex.handle_error(
                925, ['asset_type', 'update_asset', 'asset_type', 'asset_type', asset_type]
            )

        return self.tc_requests.victim_get_asset(self.unique_id, asset_type, asset_id)

    def get_email_asset(self, asset_id):
        """Retrieve an email asset of the Victim

        Args:
            asset_id: (int) the id of the asset being retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'EMAIL')

    def get_network_asset(self, asset_id):
        """Retrieve an network asset of the Victim

        Args:
            asset_id: (int) the id of the asset being retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'NETWORK')

    def get_phone_asset(self, asset_id):
        """Retrieve an phone asset of the Victim

        Args:
            asset_id: (int) the id of the asset being retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'PHONE')

    def get_social_asset(self, asset_id):
        """Retrieve an social asset of the Victim

        Args:
            asset_id: (int) the id of the asset being retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'SOCIAL')

    def get_web_asset(self, asset_id):
        """Retrieve an web asset of the Victim

        Args:
            asset_id: (int) the id of the asset being retrieved.

        Return:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'WEB')

    @staticmethod
    def is_victim():
        """Return True if object is a victim."""
        return True

    @property
    def name(self):
        """Return the Victim name."""
        return self._data.get('name')

    @name.setter
    def name(self, name):
        """Set the Victim name."""
        self._data['name'] = name

    def network_assets(self):
        """
        Gets the network assets of a Victim

        Yield:
            Json: The asset being retrieved.
        """
        return self.assets(asset_type='NETWORK')

    def social_assets(self):
        """
        Gets the social assets of a Victim

        Yield:
            Json: The asset being retrieved.
        """
        return self.assets(asset_type='SOCIAL')

    def phone_assets(self):
        """
        Gets the phone assets of a Victim

        Yield:
            Json: The asset being retrieved.
        """
        return self.assets(asset_type='PHONE')

    def update_asset(self, asset_type, asset_id, body=None):
        """
        Updates a asset of a Victim

        Valid asset_type and optional identifier:
        + email
        + network
        + phone
        + social
        + web

        Args:
            asset_id: (int) the id of the asset being deleted.
            asset_type: (str) The type of asset to be retrieved.
            body: (dict) the body of the asset being updated.

        Return:
            requests.Response: The response from the API call.
        """
        if body is None:
            body = {}

        if asset_type is None:
            self._tcex.handle_error(
                925, ['asset_type', 'update_asset', 'asset_type', 'asset_type', asset_type]
            )

        return self.tc_requests.victim_update_asset(self.unique_id, asset_type, asset_id, body)

    def update_email_asset(self, asset_id, address=None, address_type=None):
        """Update a email asset of the Victim

        Args:
            asset_id: (int) the id of the asset being updated.
            address: (str) The asset address.
            address_type: (str) The asset address type

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {}
        if address:
            asset_data['address'] = address
        if address_type:
            asset_data['addressType'] = address_type
        return self.update_asset('EMAIL', asset_id, asset_data)

    def update_network_asset(self, asset_id, account=None, network=None):
        """Update a network asset of the Victim

        Args:
            asset_id: (int) the id of the asset being updated.
            account: (str) The asset account.
            network: (str) The asset network

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {}
        if account:
            asset_data['account'] = account
        if network:
            asset_data['network'] = network
        return self.update_asset('NETWORK', asset_id, asset_data)

    def update_phone_asset(self, asset_id, phone_type=None):
        """Update a phone asset of the Victim

        Args:
            asset_id: (int) the id of the asset being updated.
            phone_type: (str) The phone type account.

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {}
        if phone_type:
            asset_data['phoneType'] = phone_type
        return self.update_asset('PHONE', asset_id, asset_data)

    def update_social_asset(self, asset_id, account=None, network=None):
        """Update a social asset of the Victim

        Args:
            asset_id: (int) the id of the asset being updated.
            account: (str) The asset account.
            network: (str) The asset network

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {}
        if account:
            asset_data['account'] = account
        if network:
            asset_data['network'] = network
        return self.update_asset('SOCIAL', asset_id, asset_data)

    def update_web_asset(self, asset_id, web_site=None):
        """Update a web asset of the Victim

        Args:
            asset_id: (int) the id of the asset being updated.
            web_site: (str) The asset web_site.

        Return:
            requests.Response: The response from the API call.
        """
        asset_data = {}
        if web_site:
            asset_data['webSite'] = web_site
        return self.update_asset('WEB', asset_id, asset_data)

    def web_assets(self):
        """
        Gets the web assets of a Victim

        Yield:
            Json: The asset being retrieved.
        """
        return self.assets(asset_type='WEB')



"""ThreatConnect TI File"""
from ..indicator import Indicator


class File(Indicator):
    """Unique API calls for File API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            tcex (Tcex): A instance of TCEX
            md5 (str, optional): The md5 value for this Indicator.
            sha1 (str, optional): The sha1 value for this Indicator.
            sha256 (str, optional): The sha256 value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            size (str, kwargs): The file size for this Indicator.
        """
        super().__init__(tcex, sub_type='File', api_entity='file', api_branch='files', **kwargs)

        self.unique_id = (
            kwargs.get('unique_id')
            or kwargs.get('md5', None)
            or kwargs.get('sha1', None)
            or kwargs.get('sha256', None)
        )
        self.data['md5'] = self.data.get('md5') or self._hash_from_unique_id(hash_type='md5')
        self.data['sha1'] = self.data.get('sha1') or self._hash_from_unique_id(hash_type='sha1')
        self.data['sha256'] = self.data.get('sha256') or self._hash_from_unique_id(
            hash_type='sha256'
        )
        if self.unique_id:
            for value in self.unique_id.split(':'):
                value = value.strip()
                if value:
                    self.unique_id = value
                    break
        self.data['size'] = kwargs.get('size', 0)

    def _hash_from_unique_id(self, hash_type='md5'):
        if not self.unique_id:
            return None

        for hash_value in self.unique_id.split(':'):
            hash_value = hash_value.strip()
            if hash_type.lower() == 'md5' and len(hash_value) == 32:
                return hash_value
            if hash_type.lower() == 'sha1' and len(hash_value) == 40:
                return hash_value
            if hash_type.lower() == 'sha256' and len(hash_value) == 64:
                return hash_value
        return None

    def can_create(self):
        """Return True if file can be create.

        If the md5/sha1/sha256 has been provided returns that the File can be
        created, otherwise returns that the File cannot be created.
        """
        if (
            self.unique_id
            or self.data.get('md5')
            or self.data.get('sha1')
            or self.data.get('sha256')
        ):
            return True
        return False

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response.

        Args:
            json_response: The json dict to retrieve the unique_id from.
        """
        self.unique_id = (
            json_response.get('md5') or json_response.get('sha1') or json_response.get('sha256')
        )

    @staticmethod
    def build_summary(val1=None, val2=None, val3=None):
        """Construct the summary given a md5, sha1, and sha256

        Args:
            val1: md5
            val2: sha1
            val3: sha256
        """
        summary = []
        if val1 is not None:
            summary.append(val1)
        if val2 is not None:
            summary.append(val2)
        if val3 is not None:
            summary.append(val3)
        if not summary:
            # Indicator object has no logger to output warning
            pass
        return ' : '.join(summary)

    def occurrences(self):
        """Yield all occurrences that file has."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        yield from self.tc_requests.file_occurrences(
            self.api_type, self.api_branch, self.unique_id, self.owner
        )

    def occurrence(self, occurrence_id):
        """Get a file occurrence given a occurrence id."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.file_occurrence(
            self.api_type, self.api_branch, self.unique_id, occurrence_id, self.owner
        )

    def add_occurrence(self, name, date, path):
        """Add  a occurrence to the file."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.add_file_occurrence(
            self.api_type, self.api_branch, self.unique_id, name, date, path, self.owner
        )

    def delete_occurrence(self, occurrence_id):
        """Delete a file occurrence given a occurrence id."""
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.delete_file_occurrence(
            self.api_type, self.api_branch, self.unique_id, occurrence_id, self.owner
        )

    def add_action(self, action, target):
        """Validate that the provided action matches the defined validation type

            + Drop
            + Traffic
            + Archive
            + registryKey
            + userAgent
            + dnsQuery

        Args:
            action (str, kwargs): The type of action to add.
            target (bool, kwargs): The Item to add.

        Return:
            Response: The response from the API call
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.add_file_action(
            self.api_type, self.api_branch, self.unique_id, action, target, self.owner
        )

    def actions(self, action, target):
        """Validate that the provided action matches the defined validation type

            + Drop
            + Traffic
            + Archive
            + registryKey
            + userAgent
            + dnsQuery

        Args:
            action (str, kwargs): The type of action to retrieve.
            target (bool, kwargs): The Item to retrieve.

        Return:
            Yield: Yield a dict of data of the File Action
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.get_file_actions(
            self.api_type, self.api_branch, self.unique_id, action, target, self.owner
        )

    def delete_action(self, action, target):
        """Validate that the provided action matches the defined validation type

            + Drop
            + Traffic
            + Archive
            + registryKey
            + userAgent
            + dnsQuery

        Args:
            action (str, kwargs): The type of action to delete.
            target (bool, kwargs): The Item to delete.

        Return:
            Response: The response from the API call
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        return self.tc_requests.delete_file_action(
            self.api_type, self.api_branch, self.unique_id, action, target, self.owner
        )



"""ThreatConnect TI Email"""
from ..group import Group


class Email(Group):
    """Unique API calls for Email API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        body (str): The body for this Email.
        from_addr (str, kwargs): The **from** address for this Email.
        header (str): The header for this Email.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        subject (str): The subject for this Email.
        to (str, kwargs): The **to** address for this Email.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties."""
        super().__init__(tcex, sub_type='Email', api_entity='email', api_branch='emails', **kwargs)



"""ThreatConnect TI Adversary"""
from ..group import Group


class Adversary(Group):
    """Unique API calls for Adversary API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(
            tcex, sub_type='Adversary', api_entity='adversary', api_branch='adversaries', **kwargs
        )

    def add_asset(self, asset_type, asset_value):
        """Add an asset to the Adversary

        Args:
            asset_type: (str) Either phone, handle, or urL
            asset_value: (str) the value for the asset

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        asset_methods = {
            'handle': self.tc_requests.add_adversary_handle_asset,
            'phone': self.tc_requests.add_adversary_phone_asset,
            'url': self.tc_requests.add_adversary_url_asset,
        }

        # handle invalid input
        if asset_methods.get(asset_type.lower()) is None:
            self._tcex.handle_error(
                925, ['asset_type', 'assets', 'asset_type', 'asset_type', asset_type]
            )

        return asset_methods.get(asset_type.lower())(self.unique_id, asset_value)

    def add_handle_asset(self, value):
        """Add a Handle asset to the adversary.

        Args:
            value: The value of the asset

        Returns:
            requests.Response: The response from the API call.
        """
        return self.add_asset('HANDLE', value)

    def add_phone_asset(self, value):
        """Add a phone asset to the adversary.

        Args:
            value: The value of the asset

        Returns:
            requests.Response: The response from the API call.
        """
        return self.add_asset('PHONE', value)

    def add_url_asset(self, value):
        """Add a URL asset to the adversary.

        Args:
            value: The value of the asset

        Returns:
            requests.Response: The response from the API call.
        """
        return self.add_asset('URL', value)

    def asset(self, asset_id, asset_type, action='GET'):
        """Get specific Adversary asset type from API

        Args:
            asset_id: (str) The ID of the asset.
            asset_type: (str) Either phone, handle, or url.
            action: (str): The HTTP method (e.g., DELETE or GET)

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        asset_methods = {
            'handle': self.tc_requests.adversary_handle_asset,
            'phone': self.tc_requests.adversary_phone_asset,
            'url': self.tc_requests.adversary_url_asset,
        }

        # handle invalid input
        if asset_methods.get(asset_type.lower()) is None:
            self._tcex.handle_error(
                925, ['asset_type', 'assets', 'asset_type', 'asset_type', asset_type]
            )

        return asset_methods.get(asset_type.lower())(self.unique_id, asset_id, action=action)

    def assets(self, asset_type=None):
        """Retrieve all of the assets of a given asset_type

        Args:
            asset_type: (str) Either None, PHONE, HANDLE, or URL

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        asset_methods = {
            'handle': self.tc_requests.adversary_handle_assets,
            'phone': self.tc_requests.adversary_phone_assets,
            'url': self.tc_requests.adversary_url_assets,
        }

        if asset_type is None:
            return self.tc_requests.adversary_assets(self.unique_id)

        # handle invalid input
        if asset_methods.get(asset_type.lower()) is None:
            self._tcex.handle_error(
                925, ['asset_type', 'assets', 'asset_type', 'asset_type', asset_type]
            )

        return asset_methods.get(asset_type.lower())(self.unique_id)

    def delete_asset(self, asset_id, asset_type):
        """Delete the asset with the provided asset_id.

        Args:
            asset_id: The id of the asset.
            asset_type: The asset type.

        Returns:
            requests.Response: The response from the API call.
        """
        return self.asset(asset_id, asset_type=asset_type, action='DELETE')

    def delete_handle_asset(self, asset_id):
        """Delete the handle asset with the passed in id

        Args:
            asset_id: The id of the asset to be deleted

        Returns:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'HANDLE')

    def delete_phone_asset(self, asset_id):
        """Delete the phone asset with the passed in id

        Args:
            asset_id: The id of the asset to be deleted

        Returns:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'PHONE')

    def delete_url_asset(self, asset_id):
        """Delete the url asset with the passed in id

        Args:
            asset_id: The id of the asset to be deleted

        Returns:
            requests.Response: The response from the API call.
        """
        return self.delete_asset(asset_id, 'URL')

    def get_asset(self, asset_id, asset_type):
        """Get the asset with the provided asset_id & asset_type.

        Args:
            asset_id: (str) The ID of the asset.
            asset_type: (str) Either None, PHONE, HANDLE, or URL

        Returns:
            requests.Response: The response from the API call.
        """
        return self.asset(asset_id, asset_type=asset_type)

    def get_handle_asset(self, asset_id):
        """Get the handle asset with the passed in id

        Args:
            asset_id: The id of the asset.

        Returns:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'HANDLE')

    def get_phone_asset(self, asset_id):
        """Get the phone asset with the passed in id

        Args:
            asset_id: The id of the asset.

        Returns:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'PHONE')

    def get_url_asset(self, asset_id):
        """Get the url asset with the passed in id

        Args:
            asset_id: The id of the asset.

        Returns:
            requests.Response: The response from the API call.
        """
        return self.get_asset(asset_id, 'URL')

    # def handle_asset(self, asset_id, action='GET'):
    #     """Get the handle asset with the passed in id.

    #     Args:
    #         asset_id: The id of the asset.
    #         action: (str): The HTTP method (e.g., DELETE or GET)

    #     Returns:
    #         requests.Response: The response from the API call.
    #     """
    #     return self.asset(asset_id, 'HANDLE', action=action)

    def handle_assets(self):
        """Return all of the handle assets"""
        return self.assets(asset_type='HANDLE')

    # def phone_asset(self, asset_id, action='GET'):
    #     """Get the phone asset with the passed in id.

    #     Args:
    #         asset_id: The id of the asset.
    #         action: (str): The HTTP method (e.g., DELETE or GET)

    #     Returns:
    #         requests.Response: The response from the API call.
    #     """
    #     return self.asset(asset_id, 'PHONE', action=action)

    def phone_assets(self):
        """Return all of the phone assets"""
        return self.assets(asset_type='PHONE')

    # def url_asset(self, asset_id, action='GET'):
    #     """Get the url asset with the passed in id.

    #     Args:
    #         asset_id: The id of the asset.
    #         action: (str): The HTTP method (e.g., DELETE or GET)

    #     Returns:
    #         requests.Response: The response from the API call.
    #     """
    #     return self.asset(asset_id, 'URL', action=action)

    def url_assets(self):
        """Return all of the url assets"""
        return self.assets(asset_type='URL')



"""ThreatConnect Assignee Module"""


class User:
    """User Object used for assignee and createdBy.

    Args:
        first_name (str, kwargs): The first name of the user.
        id (id, kwargs): The id of the user.
        last_name (str, kwargs): The last name of the user.
        pseudonym (str, kwargs): The pseudonym of the user.
        role (str, kwargs): The role of the user.
        user_name (str, kwargs): The user name of the user.
    """

    def __init__(self, **kwargs):
        """Initialize Class properties."""
        self._transform_kwargs(kwargs)
        self._first_name = kwargs.get('first_name', None)
        self._id = kwargs.get('id', None)
        self._last_name = kwargs.get('last_name', None)
        self._pseudonym = kwargs.get('pseudonym', None)
        self._role = kwargs.get('role', None)
        self._user_name = kwargs.get('user_name', None)

    def __str__(self):
        """Printable version of Object"""
        printable_string = ''
        for key, value in sorted(vars(self).items()):
            key = f'{key.lstrip("_")} '
            if value is None:
                printable_string += f'{key:.<40} {"null":<50}\n'
            elif isinstance(value, (int, str)):
                printable_string += f'{key:.<40} {value:<50}\n'
            else:
                printable_string += f'{key:.<40} {"<object>":<50}\n'
        return printable_string

    @property
    def _metadata_map(self):
        """Return a mapping of kwargs to expected args."""
        return {
            'dateAdded': 'date_added',
            'firstName': 'first_name',
            'lastName': 'last_name',
            'userName': 'user_name',
        }

    def _transform_kwargs(self, kwargs):
        """Map the provided kwargs to expected arguments."""
        for key in dict(kwargs):
            new_key = self._metadata_map.get(key, key)
            kwargs[new_key] = kwargs.pop(key)

    @property
    def body(self):
        """Return a dict representation of the Creator class."""
        return self.as_dict

    @property
    def as_dict(self):
        """Return a dict representation of the Creator class."""
        properties = vars(self)
        as_dict = {}
        for key, value in properties.items():
            key = key.lstrip('_')
            if value is None:
                continue
            as_dict[key] = value

        if not as_dict:
            return None

        return as_dict

    @property
    def as_entity(self):
        """Return the USER as a entity."""
        return {'type': 'User', 'value': self.user_name, 'id': self.id}

    @property
    def first_name(self):
        """Return the First Name for the Case Creator."""
        return self._first_name

    @first_name.setter
    def first_name(self, first_name):
        """Set the First Name for the Case Creator."""
        self._first_name = first_name

    @property
    def id(self):
        """Return the ID for the Case Creator."""
        return self._id

    @id.setter
    def id(self, creator_id):
        """Set the ID for the Case Creator."""
        self._id = creator_id

    @property
    def last_name(self):
        """Return the **Last Name** for the User Object."""
        return self._last_name

    @last_name.setter
    def last_name(self, last_name):
        """Set the **Last Name** for the User Object."""
        self._last_name = last_name

    @property
    def pseudonym(self):
        """Return the Pseudonym for the Case Creator."""
        return self._pseudonym

    @pseudonym.setter
    def pseudonym(self, pseudonym):
        """Set the Pseudonym for the Case Creator."""
        self._pseudonym = pseudonym

    @property
    def role(self):
        """Return the Role for the Case Creator."""
        return self._role

    @role.setter
    def role(self, role):
        """Set the Role for the Case Creator."""
        self._role = role

    @property
    def user_name(self):
        """Return the User Name for the Case Creator."""
        return self._user_name

    @user_name.setter
    def user_name(self, user_name):
        """Set the User Name for the Case Creator."""
        self._user_name = user_name


class Users:
    """Sub class of the Cases object. Used to map the users to.

    Args:
        users (list): A array of user data
    """

    def __init__(self, users=None):
        """Initialize Class properties."""
        if users is None:
            users = []
        self._users = []

        for user in users:
            if isinstance(user, User):
                self._users.append(user)
            else:
                self._users.append(User(**user))

    @property
    def users(self):
        """Return the **Users**."""
        return self._users

    @users.setter
    def users(self, users):
        """Set the **Users**."""
        self._users = users

    @property
    def as_dict(self):
        """Return a dict representation of the UsersData class."""
        data = []
        for user in self.users:
            data.append(user.as_dict)

        return {'data': data}

    @property
    def body(self):
        """Return a body representation of the Creator class."""
        body = []
        for user in self.users:
            body.append(user.body)

        return {'data': body}


class Assignee(User):
    """Assignee Object for Case Management.

    For User type the user_name or id fields is required and
    for the Group type the name or id is required.

    Args:
        first_name (str, kwargs): The first name of the User.
        id (id, kwargs): The id of the User.
        name (str, kwargs): The name of the Group.
        last_name (str, kwargs): The last name of the user.
        pseudonym (str, kwargs): The pseudonym of the User.
        role (str, kwargs): The role of the User.
        user_name (str, kwargs): The user name of the User.
        type (str, kwargs): The assignee type. Default to User.
    """

    def __init__(self, type='User', **kwargs):  # pylint: disable=redefined-builtin
        """Initialize Class properties."""
        self._name = kwargs.get('name')
        self._type = type
        super().__init__(**kwargs)

    @property
    def as_dict(self):
        """Return a dict representation of the Assignee class."""
        user_data = super().as_dict
        del user_data['type']
        data = {
            'data': user_data,
            'type': self._type,
        }
        if self._name:
            data['data']['name'] = self._name
        return data



"""Package version and other metadata"""
__author__ = 'ThreatConnect (support@threatconnect.com)'
__author_email__ = 'support@threatconnect.com'
__description__ = 'ThreatConnect Exchange App Framework'
__license__ = 'Apache License, Version 2'
__package_name__ = 'tcex'
__url__ = 'https://github.com/ThreatConnect-Inc/tcex'
__version__ = '2.0.13'
__download_url__ = f'https://github.com/ThreatConnect-Inc/tcex/tarball/{__version__}'



"""TcEx testing profile Class."""
# standard library
import json
import math

# import os
import re
import sys
from base64 import b64encode
from typing import Optional, Union

# third-party
import colorama as c

# autoreset colorama
c.init(autoreset=True, strip=False)


class Interactive:
    """Testing Profile Interactive Class."""

    def __init__(self, profile: object):
        """Initialize Class properties.

        Args:
            profile (Profile): The profile object to build interactive inputs.
        """
        self.profile = profile

        # properties
        self._inputs = {
            'optional': {},
            'required': {},
        }
        self._no_selection_text = 'No Selection'
        self._staging_data = {'kvstore': {}}
        # self._user_defaults = None
        self.collect_type_map = {
            'Any': self.collect_string,
            'Binary': self.collect_binary,
            'BinaryArray': self.collect_binary_array,
            'KeyValue': self.collect_key_value,
            'KeyValueArray': self.collect_key_value_array,
            'String': self.collect_string,
            'StringArray': self.collect_string_array,
            'TCEntity': self.collect_tcentity,
            'TCEntityArray': self.collect_tcentity_array,
        }
        self.exit_codes = []
        self.input_type_map = {
            'boolean': self.present_boolean,
            'choice': self.present_choice,
            'keyvaluelist': self.present_key_value_list,
            'multichoice': self.present_multichoice,
            'string': self.present_string,
        }
        # self.user_defaults_filename = os.path.join('tests', '.user_defaults')

    def _default(self, data: dict) -> Union[list, str]:  # pylint: disable=unused-argument
        """Return the best option for default.

        Args:
            data: The install.json params object.

        Returns:
            list, str: The default value for the input.
        """
        if data.get('type').lower() == 'boolean':
            default = str(data.get('default', 'false')).lower()
        elif data.get('type').lower() == 'choice':
            default = 0
            valid_values: list = self._expand_valid_values(data.get('validValues', []))
            if data.get('name') == 'tc_action':
                for vv in valid_values:
                    if self.profile.feature.lower() == vv.replace(' ', '_').lower():
                        default = vv
                        break
            else:
                default: str = data.get('default')
        elif data.get('type').lower() == 'multichoice':
            default: str = data.get('default')
            if default is not None and isinstance(default, str):
                default: list = default.split('|')
        else:
            default = data.get('default')
            # if default is None:
            #     # set default from user default file
            #     default = self.user_defaults.get(data.get('name'))
        return default

    def _expand_valid_values(self, valid_values: list) -> list:
        """Expand supported playbook variables to their full list.

        Args:
            valid_values (list): The list of valid values for Choice or MultiChoice inputs.

        Returns:
            list: An expanded list of valid values for Choice or MultiChoice inputs.
        """
        valid_values = list(valid_values)
        if '${ARTIFACT_TYPES}' in valid_values:
            valid_values.remove('${ARTIFACT_TYPES}')
            valid_values.extend(
                [
                    'ASN',
                    'Asset Group ID',
                    'Certificate File',
                    'CIDR',
                    'Credential ID',
                    'Document Metadata',
                    'Email Address',
                    'Email Attachment File',
                    'Email Attachment File Name',
                    'Email Body',
                    'Email Message File',
                    'Email Subject',
                    'Event File',
                    'Exploit ID',
                    'File Hash',
                    'Filter ID',
                    'Hashtag',
                    'Host',
                    'Image File',
                    'IP Address',
                    'Log File',
                    'MutEx',
                    'PCAP File',
                    'Policy ID',
                    'Registry Key',
                    'Results ID',
                    'Screenshot File',
                    'Tactic ID',
                    'Technique ID',
                    'Ticket ID',
                    'Timestamp',
                    'URL',
                    'User Agent',
                    'Vulnerability Detection ID',
                    'Vulnerability ID',
                ]
            )
        elif '${GROUP_TYPES}' in valid_values:
            valid_values.remove('${GROUP_TYPES}')
            valid_values.extend(
                [
                    'Adversary',
                    'Campaign',
                    'Document',
                    'Email',
                    'Event',
                    'Incident',
                    'Intrusion Set',
                    'Signature',
                    'Task',
                    'Threat',
                ]
            )
        elif '${INDICATOR_TYPES}' in valid_values:
            valid_values.remove('${INDICATOR_TYPES}')
            r = self.profile.session.get('/v2/types/indicatorTypes')
            if r.ok:
                valid_values.extend(
                    [t.get('name') for t in r.json().get('data', {}).get('indicatorType', {})]
                )
        elif '${OWNERS}' in valid_values:
            valid_values.remove('${OWNERS}')
            r = self.profile.session.get('/v2/owners')
            if r.ok:
                valid_values.extend(
                    [o.get('name') for o in r.json().get('data', {}).get('owner', {})]
                )
        elif '${USERS}' in valid_values:
            valid_values.remove('${USERS}')
            r = self.profile.session.get('/v2/owners/mine/members')
            if r.ok:
                valid_values.extend(
                    [o.get('userName') for o in r.json().get('data', {}).get('user', {})]
                )
        elif '${USER_GROUPS}' in valid_values:
            valid_values.remove('${USER_GROUPS}')
            valid_values.extend(['User Group 1', 'User Group 1'])

        return valid_values

    def _input_value(self, label: str, option_text: Optional[str] = None) -> str:
        """Return user input.

        Args:
            label: The label to display to the user.
            option_text: the Option text to display to the user.

        Returns:
            str: The value selected by the user.
        """
        # update option text to include help message
        option_text = option_text or ''
        if option_text:
            # add space for cleaness in user display
            option_text = f' {option_text}'

        print(f'{c.Fore.WHITE}[? for help]')
        prompt = f'{c.Fore.MAGENTA}{label}{c.Fore.RESET}{c.Style.BRIGHT}{option_text}: '
        input_value = input(prompt).strip()  # nosec

        # handle special user inputs
        if input_value == '?':
            self.present_help()
            return self._input_value(label, option_text)

        return input_value

    @staticmethod
    def _split_list(data: list) -> tuple:
        """Split a list in two "equal" parts.

        Args:
            data: The list of data to split into two equal parts.

        Returns:
            tuple: The two halves of the list.
        """
        half: int = math.ceil(len(data) / 2)
        return data[:half], data[half:]

    def add_input(self, name: str, data: dict, value: str) -> None:
        """Add an input to inputs.

        Args:
            name: The name of the input.
            data: The install.json params object.
            value: The value for the input.
        """
        if data.get('required', False):
            self._inputs['required'].setdefault(name, value)
        else:
            self._inputs['optional'].setdefault(name, value)

    # def add_user_default(self, key, value, data_type=None):
    #     """Add data to user default."""
    #     self.user_defaults.setdefault(self.profile.feature, {})
    #     if data_type is None:
    #         self.user_defaults[self.profile.feature][key] = value
    #     else:
    #         # store the value under the appropriate data type
    #         self.user_defaults[self.profile.feature].setdefault(key, {})
    #         self.user_defaults[self.profile.feature][key].setdefault(data_type, value)

    #     if self.user_defaults.get('base') is None:
    #         self.user_defaults['base'] = self.user_defaults[self.profile.feature]

    def add_staging_data(self, name: str, type_: str, value: str) -> str:
        """Create staging data and return variable value.

        Args:
            name: The name of the input.
            type_: The type of input (Binary, StringArray, etc.)
            value: The value to write in the staging data.

        Returns:
            str: The newly create variable string.
        """
        arg_value = value
        if (
            self.profile.ij.runtime_level.lower() not in ['triggerservice', 'webhooktriggerservice']
            and value is not None
        ):
            arg_value: str = self.profile.ij.create_variable(name, type_)
            self._staging_data['kvstore'].setdefault(arg_value, value)

        return arg_value

    def collect_binary(self, **kwargs) -> str:
        """Collect binary data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            feedback (bool, kwargs): If True user feedback will be printed.
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            str: The input str from the user.
        """
        input_value: str = self._input_value('Input', kwargs.get('option_text'))
        if not input_value:
            # if no default value and required force user to input again
            if kwargs.get('default') is None and kwargs.get('required') is True:
                self.print_required()
                return self.collect_binary(**kwargs)

        if input_value not in [None, '']:
            input_data: str = b64encode(input_value.encode()).decode()
            feedback = f'{input_value} -> ({input_data})'
        else:
            input_data: str = kwargs.get('default')
            feedback = input_data

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(feedback)

        return input_data

    def collect_binary_array(self, **kwargs) -> list:
        """Collect binary array data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            list: The input list from the user.
        """
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_binary(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False  # only the first input is required

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_boolean(self, **kwargs) -> bool:
        """Collect binary data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            option_text (str, kwargs): The text shown to the user.

        Returns:
            bool: The boolean value select by the user.
        """
        input_value = self._input_value('Input', kwargs.get('option_text'))
        if input_value == '':
            input_value = kwargs.get('default')

        if str(input_value).lower() not in ['0', 'f', 'false', '1', 't', 'true']:
            self.print_invalid_bool()
            return self.collect_boolean(**kwargs)

        # covert input value to a proper boolean
        input_value = self.profile.utils.to_bool(input_value)

        # print user feedback
        self.print_feedback(input_value)

        return input_value

    def collect_choice(self, **kwargs) -> str:
        """Collect choice data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.
            valid_values (str, kwargs): A list of valid values

        Returns:
            str: The users selected choice.
        """
        # collect input value from user and set default if required
        input_value: str = self._input_value('Choice', kwargs.get('option_text')) or kwargs.get(
            'default'
        )

        # ensure input value is provided when input is required
        if input_value is None and kwargs.get('required') is True:
            self.print_required()
            return self.collect_choice(**kwargs)

        # if input value is None then there is not need to continue
        if input_value is None:
            return input_value

        # set valid values
        valid_values: list = kwargs.get('valid_values', [])

        # convert to int or recollect input
        try:
            input_value = int(input_value)
        except ValueError:
            self.print_invalid_index(f'0-{len(valid_values)}')
            return self.collect_choice(**kwargs)

        # ensure input value is valid
        valid_index_values = [i for i, _ in enumerate(valid_values)]
        # valid_index_values = list(range(0, len(valid_values) - 1))
        if input_value not in valid_index_values:
            self.print_invalid_index(f'0-{len(valid_values)}')
            return self.collect_choice(**kwargs)

        # using index value provided by user, set value to valid value
        input_value = valid_values[input_value]
        if input_value == self._no_selection_text:
            # special case for when user select no selection
            input_value = None

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_exit_code(self, **kwargs) -> int:
        """Collect exit codes.

        Args:
            option_text (str, kwargs): The text shown to the user.

        Returns:
            str: The users provided exit code.
        """
        input_value = self._input_value('Code', kwargs.get('option_text'))

        if input_value != '':
            try:
                input_value = int(input_value)
            except ValueError:
                self.print_invalid_exit_code()
                return self.collect_exit_code(**kwargs)

            if input_value not in [0, 1, 3]:
                self.print_invalid_exit_code()
                return self.collect_exit_code(**kwargs)

        return input_value

    def collect_exit_codes(self, **kwargs) -> list:
        """Collect exit codes.

        Returns:
            list: The users provided exit code.
        """
        input_values = []
        while True:
            input_value = self.collect_exit_code(**kwargs)
            if input_value == '':
                break
            input_values.append(input_value)

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = [0]

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_key_value(self, **kwargs) -> dict:
        """Collect key value data.

        Args:
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            dict: The users provided key value input.
        """
        input_value = None
        key = self._input_value('Key', option_text=kwargs.get('option_text'))

        # ensure input value is provided when input is required
        if key == '' and kwargs.get('required') is True:
            self.print_required()
            return self.collect_key_value(**kwargs)

        if key != '':
            value = self._input_value('Value')
            input_value = {'key': key, 'value': value}
        else:
            input_value = kwargs.get('default')

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_key_value_array(self, **kwargs) -> list:
        """Collect key value array data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            list: The users provided list of key value inputs.
        """
        input_values = []
        required: bool = kwargs.get('required')
        while True:
            input_value = self.collect_key_value(
                default=kwargs.get('default'),
                feedback=False,
                option_test=kwargs.get('option_text'),
                required=required,
            )
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_multichoice(self, **kwargs) -> list:
        """Collect multichoice data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.
            valid_values (str, kwargs): A list of valid values

        Returns:
            list: The users provided list of choice inputs.
        """
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_choice(
                feedback=False,
                # option_text=kwargs.get('option_text'),
                required=required,
                valid_values=kwargs.get('valid_values'),
            )
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        input_values = list(set(input_values))
        if input_values:
            # format multichoice value as pipe delimited string
            input_values = '|'.join(input_values)
        else:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_string(self, **kwargs) -> str:
        """Collect string data

        Args:
            option_text (str, kwargs): The text shown to the user.
            default (str, kwargs): The default value if no value provided by user.

        Returns:
            str: The user provided input.
        """
        input_value = self._input_value('Input', kwargs.get('option_text', ''))
        if not input_value:
            input_value = kwargs.get('default')

        if input_value is None and kwargs.get('required', False) is True:
            self.print_required()
            return self.collect_string(**kwargs)

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        # APP-622 - handle null/None values
        if input_value == 'null':
            input_value = None
        elif input_value in ['"null"', "'null'"]:
            input_value = 'null'

        return input_value

    def collect_string_array(self, **kwargs) -> list:
        """Collect string data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            str: The user provided input.
        """
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_string(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_tcentity(self, **kwargs) -> dict:
        """Collect tcentity data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            str: The user provided input.
        """
        input_value = None
        id_ = self._input_value('ID')
        if id_:
            value = self._input_value('Value')
            type_ = self._input_value('Type')
            input_value = {'id': id_, 'value': value, 'type': type_}

        if input_value is None and kwargs.get('required', False) is True:
            self.print_required()
            return self.collect_tcentity(**kwargs)

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_tcentity_array(self, **kwargs) -> list:
        """Collect tcentity array data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.

        Returns:
            list: The user provided inputs.
        """
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_tcentity(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    @property
    def inputs(self) -> dict:
        """Return inputs dict."""
        return self._inputs

    def present(self) -> None:
        """Present interactive menu to build profile."""

        def params_data() -> tuple:
            # handle non-layout and layout based App appropriately
            if self.profile.lj.has_layout:
                # using inputs from layout.json since they are required to be in order
                # (display field can only use inputs previously defined)
                for name in self.profile.lj.params_dict:
                    # get data from install.json based on name
                    data = self.profile.ij.params_dict.get(name)
                    yield name, data

                # hidden fields will not be in layout.json so they need to be include manually
                for name, data in self.profile.ij.filter_params_dict(hidden=True).items():
                    yield name, data
            else:
                for name, data in self.profile.ij.params_dict.items():
                    yield name, data

        inputs = {}
        for name, data in params_data():
            if data.get('serviceConfig'):
                # inputs that are serviceConfig are not applicable for profiles
                continue

            if not data.get('hidden'):
                # each input will be checked for permutations if the App has layout and not hidden
                if not self.profile.permutations.validate_input_variable(name, inputs):
                    continue

            # present the input
            value: str = self.input_type_map.get(data.get('type').lower())(name, data)

            # update inputs
            inputs[name] = value

        self.present_exit_code()

    def present_boolean(self, name: str, data) -> bool:
        """Build a question for boolean input.

        Args:
            name: The name of the input field.
            data: The install.json input param object.

        Returns:
            bool: The user provided input.
        """
        # print header information
        self.print_header(data)

        default = self._default(data)
        valid_values = ['true', 'false']

        option_default = 'false'
        option_text = ''
        options = []
        for v in valid_values:
            if v.lower() == default.lower():
                option_default = v
                v = f'[{v}]'
            options.append(v)
        option_text = f'''({'/'.join(options)})'''

        value = self.collect_boolean(default=option_default, option_text=option_text)

        # add input
        self.add_input(name, data, value)

        return value

    def present_choice(self, name: str, data: dict) -> str:
        """Build a question for choice input.

        Args:
            name: The name of the input field.
            data: The install.json input param object.

        Returns:
            str: The user provided input.
        """
        # print header information
        self.print_header(data)

        default = self._default(data)
        option_index = 0
        valid_values = self._expand_valid_values(data.get('validValues', []))
        if data.get('required', False) is False:
            # add option to invalidate defaults
            valid_values.insert(0, self._no_selection_text)

        # default value needs to be converted to index
        if default:
            try:
                option_index = valid_values.index(default)
            except ValueError:
                # if "magic" variable (e.g., ${GROUP_TYPES}) was not expanded then use index 0.
                # there is no way to tell if the default value is be part of the expansion.
                if any([re.match(r'^\${.*}$', v) for v in valid_values]):
                    option_index = 0
                else:
                    print(
                        f'''{c.Fore.RED}Invalid value of ({default}) for {data.get('name')}, '''
                        'check that default value and validValues match in install.json.'
                    )
                    sys.exit()
        option_text = f'[{option_index}]'

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(valid_values):
            options.append(f'{i}. {v}')

        # display options list into two columns
        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        # collect user input
        value = self.collect_choice(
            default=option_index, option_text=option_text, valid_values=valid_values
        )

        # add input
        self.add_input(name, data, value)

        return value

    def present_data_types(self, data_types: list, required: Optional[bool] = False) -> str:
        """Present data types options.

        Args:
            data_types: A list of optional data types.
            required: If False the no selection option will be added.

        Returns:
            str: The user provided input.
        """
        if 'Any' in data_types:
            data_types = [
                'Binary',
                'BinaryArray',
                'KeyValue',
                'KeyValueArray',
                'String',
                'StringArray',
                'TCEntity',
                'TCEntityArray',
            ]

        # add option to not select an index value if input is not required
        if required is False:
            data_types.insert(0, self._no_selection_text)

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(data_types):
            options.append(f'{i}. {v}')

        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        index = self._input_value('Type', '[0]') or 0

        try:
            data_type = data_types[int(index)]
        except (IndexError, TypeError, ValueError):
            print(
                f'{c.Fore.RED}Invalid index of {index} provided. '
                f'Please provide a integer between 0-{len(data_types) - 1}'
            )
            sys.exit(1)

        return data_type

    def present_exit_code(self) -> None:
        """Provide user input for exit code."""
        self.print_header({'label': 'Exit Codes'})
        self.exit_codes = list(set(self.collect_exit_codes(default=[0], option_text='[0]')))

    @staticmethod
    def present_help() -> None:
        """Provide user help information."""
        print(
            f'{c.Fore.CYAN}For String type inputs: \n'
            ' * A value of null will be treated as an actual null value.\n'
            ' * Using "null" or \'null\' to insert a string of null.\n'
        )
        print(f'{c.Fore.CYAN}When done entering array data press enter to continue.')

    def present_key_value_list(self, name: str, data: dict) -> None:
        """Build a question for key value list input.

        Args:
            name: The name of the input field.
            data: The install.json input param object.

        Returns:
            str: The user provided input.
        """
        # print header information
        self.print_header(data)

        # the default value from install.json or user_data
        default = self._default(data)  # array of default values

        # collect input
        input_data = self.collect_key_value_array(default=default, required=data.get('required'))

        # create variable
        variable = self.add_staging_data(name, 'KeyValueArray', input_data)

        # add input to args
        self.add_input(name, data, variable)

        # user feedback
        feedback_data = input_data
        if input_data is not None:
            feedback_data = json.dumps(feedback_data)

        # # update default
        # if default is None:
        #     self.add_user_default(name, input_data)

        return variable

    def present_multichoice(self, name: str, data: dict) -> list:
        """Build a question for multichoice input.

        Args:
            name: The name of the input field.
            data: The install.json input param object.

        Returns:
            list: The user provided inputs.
        """
        # print header information
        self.print_header(data)

        default = self._default(data)  # array of default values
        option_indexes = [0]
        valid_values = self._expand_valid_values(data.get('validValues', []))
        if data.get('required', False) is False:
            # add option to invalidate defaults
            valid_values.insert(0, self._no_selection_text)

        # default values will be return as an array (e.g., one|two -> ['one'. 'two']).
        # using the valid values array we can look up these values to show as default in input.
        if default:
            option_indexes = []
            for d in default:
                try:
                    option_indexes.append(valid_values.index(d))
                except ValueError:
                    # if "magic" variable (e.g., ${GROUP_TYPES}) was not expanded then skip value.
                    # there is no way to tell if the default value is be part of the expansion.
                    if any([re.match(r'^\${.*}$', v) for v in valid_values]):
                        continue

                    print(
                        f'''{c.Fore.RED}Invalid value of ({d}) for {data.get('name')}, check '''
                        'that default value(s) and validValues match in install.json.'
                    )
                    sys.exit()
        option_text = f''' [{','.join([str(v) for v in option_indexes])}]'''

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(valid_values):
            options.append(f'{i}. {v}')

        # display options list into two columns
        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        # collect user input
        values = self.collect_multichoice(
            default=option_indexes,
            option_text=option_text,
            required=data.get('required'),
            valid_values=valid_values,
        )

        # add input
        self.add_input(name, data, values)

        return values

    def present_string(self, name: str, data: dict) -> str:
        """Build a question for string input.

        Args:
            name: The name of the input field.
            data: The install.json input param object.

        Returns:
            str: The user provided input.
        """
        # display header information
        self.print_header(data)

        # use playbook data types to determine what input to provide (default to String)
        data_type = data.get('playbookDataType', ['String'])[0]
        if len(data.get('playbookDataType', [])) > 1 or data_type.lower() == 'any':
            data_type = self.present_data_types(
                data.get('playbookDataType'), required=data.get('required', False)
            )

        # no need to proceed if there is not valid data type selected.
        if data_type == self._no_selection_text:
            self.add_input(name, data, None)
            self.print_feedback('null')
            return None

        # the default value from install.json or user_data
        default = self._default(data)

        option_text = ''
        if default is not None:
            option_text = f'[{default}]'

        # use data_type to properly format collection input
        input_value = self.collect_type_map[data_type](
            default=default, option_text=option_text, required=data.get('required', False)
        )

        # add staging data and get variable name
        variable = self.add_staging_data(name, data_type, input_value)

        # add input
        self.add_input(name, data, variable)

        # # update default
        # if default is None:
        #     if len(data.get('playbookDataType', [])) > 1 or data_type.lower() == 'any':
        #         # for inputs that take multiple types we need to store user default with the type
        #         self.add_user_default(name, input_value, data_type)
        #     else:
        #         self.add_user_default(name, input_value)

        return variable

    @staticmethod
    def print_feedback(feedback_value: Union[list, str]) -> None:
        """Print the value used."""
        print(f'Using value: {c.Fore.GREEN}{feedback_value}\n')

    @staticmethod
    def print_header(data: dict) -> None:
        """Enrich the header with metadata.

        Args:
            data: The install.json input param object.
        """

        def _print_metadata(title: str, value: str) -> None:
            """Print the title and value"""
            print(f'{c.Fore.CYAN}{title!s:<22}: {c.Fore.RESET}{c.Style.BRIGHT}{value}')

        label = data.get('label', 'NO LABEL')
        print(f'\n{c.Fore.GREEN}{label}')

        # type
        _print_metadata('Type', data.get('type'))

        # default
        default = data.get('default')
        if default:
            _print_metadata('Default', default)

        # note
        note = data.get('note', '')[:200]
        if note:
            _print_metadata('Note', note)

        # required
        _print_metadata('Required', str(data.get('required', False)).lower())

        # hidden
        if data.get('hidden'):
            _print_metadata('Hidden', 'true')

        # Input Types
        pbt = ','.join(data.get('playbookDataType', []))
        if pbt:
            _print_metadata('Playbook Data Types', pbt)

        vv = ','.join(data.get('validValues', []))
        if vv:
            _print_metadata('Valid Values', vv)

        print('-' * 50)

    @staticmethod
    def print_invalid_bool() -> None:
        """Print a invalid bool error."""
        print(f'{c.Fore.RED}The provided value is not a boolean value (true/false).\n')

    @staticmethod
    def print_invalid_exit_code() -> None:
        """Print a invalid exit code error."""
        print(f'{c.Fore.RED}The provided value is not a valid exit code (0, 1).\n')

    @staticmethod
    def print_invalid_index(range_: str) -> None:
        """Print a invalid index error.

        Args:
            range_: The range of possible value for choice or multichoice selections.
        """
        print(
            f'{c.Fore.RED}The provided index value is not '
            f'valid, please select a valid value between {range_}.\n'
        )

    @staticmethod
    def print_required() -> None:
        """Print a required error."""
        print(f'{c.Fore.RED}This input is required, please enter an appropriate value.\n')

    @property
    def staging_data(self) -> None:
        """Return staging data dict."""
        return self._staging_data

    # @property
    # def user_defaults(self):
    #     """Return user defaults"""
    #     if self._user_defaults is None:
    #         user_defaults = {}
    #         if os.path.isfile(self.user_defaults_filename):
    #             with open(self.user_defaults_filename, 'r') as fh:
    #                 user_defaults = json.load(fh)

    #         # use feature defaults
    #         self._user_defaults = user_defaults.get(self.profile.feature)
    #         if self._user_defaults is None:
    #             # use base defaults if not feature defaults found
    #             self._user_defaults = user_defaults.get('base', {})

    #     return self._user_defaults



"""TcEx Framework Service Common module"""
# standard library
import json
import threading
import time
import traceback
import uuid
from datetime import datetime
from typing import Callable, Optional, Union

from .mqtt_message_broker import MqttMessageBroker


class CommonService:
    """TcEx Framework Service Common module

    Shared service logic between the supported service types:
    * API Service
    * Custom Trigger Service
    * Webhook Trigger Service
    """

    def __init__(self, tcex: object):
        """Initialize the Class properties.

        Args:
            tcex: Instance of TcEx.
        """
        self.tcex = tcex

        # properties
        self._ready = False
        self._redis_client = None
        self._start_time = datetime.now()
        self.args: object = tcex.default_args
        self.configs = {}
        self.heartbeat_max_misses = 3
        self.heartbeat_sleep_time = 1
        self.heartbeat_watchdog = 0
        self.ij = tcex.ij
        self.log = tcex.log
        self.logger = tcex.logger
        self.message_broker = MqttMessageBroker(
            broker_host=self.args.tc_svc_broker_host,
            broker_port=self.args.tc_svc_broker_port,
            broker_timeout=self.args.tc_svc_broker_conn_timeout,
            broker_token=self.args.tc_svc_broker_token,
            broker_cacert=self.args.tc_svc_broker_cacert_file,
            logger=tcex.log,
        )
        self.ready = False
        self.token = tcex.token

        # config callbacks
        self.shutdown_callback = None

    def _create_logging_handler(self):
        """Create a logging handler."""
        if self.logger.handler_exist(self.thread_name):
            return

        # create trigger id logging filehandler
        self.logger.add_pattern_file_handler(
            name=self.thread_name,
            filename=f'''{datetime.today().strftime('%Y%m%d')}/{self.session_id}.log''',
            level=self.args.tc_log_level,
            path=self.args.tc_log_path,
            # uuid4 pattern for session_id
            pattern=r'^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}.log$',
            handler_key=self.session_id,
            thread_key='session_id',
        )

    def add_metric(self, label: str, value: Union[int, str]) -> None:
        """Add a metric.

        Metrics are reported in heartbeat message.

        Args:
            label: The metric label (e.g., hits) to add.
            value: The value for the metric.
        """
        self._metrics[label] = value

    @property
    def command_map(self) -> dict:
        """Return the command map for the current Service type."""
        return {
            'heartbeat': self.process_heartbeat_command,
            'loggingchange': self.process_logging_change_command,
            'shutdown': self.process_shutdown_command,
        }

    @staticmethod
    def create_session_id() -> str:  # pylint: disable=unused-argument
        """Return a uuid4 session id.

        Returns:
            str: A unique UUID string value.
        """
        return str(uuid.uuid4())

    def heartbeat(self) -> None:
        """Start heartbeat process."""
        self.service_thread(name='heartbeat', target=self.heartbeat_monitor)

    def heartbeat_monitor(self) -> None:
        """Publish heartbeat on timer."""
        self.log.info('feature=service, event=heartbeat-monitor-started')
        while True:
            if self.heartbeat_watchdog > (
                int(self.args.tc_svc_hb_timeout_seconds) / int(self.heartbeat_sleep_time)
            ):
                self.log.error(
                    'feature=service, event=missed-heartbeat, action=shutting-service-down'
                )
                self.process_shutdown_command({'reason': 'Missed heartbeat commands.'})
                break
            time.sleep(self.heartbeat_sleep_time)
            self.heartbeat_watchdog += 1

    def increment_metric(self, label: str, value: Optional[int] = 1) -> None:
        """Increment a metric if already exists.

        Args:
            label: The metric label (e.g., hits) to increment.
            value: The increment value. Defaults to 1.
        """
        if self._metrics.get(label) is not None:
            self._metrics[label] += value

    def listen(self) -> None:
        """List for message coming from broker."""
        self.message_broker.add_on_connect_callback(self.on_connect_handler)
        self.message_broker.add_on_message_callback(
            self.on_message_handler, topics=[self.args.tc_svc_server_topic]
        )
        self.message_broker.register_callbacks()

        # start listener thread
        self.service_thread(name='broker-listener', target=self.message_broker.connect)

    def loop_forever(self, sleep: Optional[int] = 1) -> bool:
        """Block and wait for shutdown.

        Args:
            sleep: The amount of time to sleep between iterations. Defaults to 1.

        Returns:
            Bool: Returns True until shutdown received.
        """
        while True:
            deadline = time.time() + sleep
            while time.time() < deadline:
                if self.message_broker.shutdown:
                    return False
                time.sleep(1)
            return True

    @property
    def metrics(self) -> dict:
        """Return current metrics."""
        # TODO: move to trigger command and handle API Service
        if self._metrics.get('Active Playbooks') is not None:
            self.update_metric('Active Playbooks', len(self.configs))
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: dict):
        """Return current metrics."""
        if isinstance(metrics, dict):
            self._metrics = metrics
        else:
            self.log.error('feature=service, event=invalid-metrics')

    def on_connect_handler(
        self, client, userdata, flags, rc  # pylint: disable=unused-argument
    ) -> None:
        """On connect method for mqtt broker."""
        self.log.info(
            f'feature=service, event=topic-subscription, topic={self.args.tc_svc_server_topic}'
        )
        self.message_broker.client.subscribe(self.args.tc_svc_server_topic)
        self.message_broker.client.disable_logger()

    def on_message_handler(
        self, client, userdata, message  # pylint: disable=unused-argument
    ) -> None:
        """On message for mqtt."""
        try:
            # messages on server topic must be json objects
            m = json.loads(message.payload)
        except ValueError:
            self.log.warning(
                f'feature=service, event=parsing-issue, message="""{message.payload}"""'
            )
            return

        # use the command to call the appropriate method defined in command_map
        command: str = m.get('command', 'invalid').lower()
        trigger_id: Optional[int] = m.get('triggerId')
        if trigger_id is not None:
            # coerce trigger_id to int in case a string was provided (testing framework)
            trigger_id = int(trigger_id)
        self.log.info(f'feature=service, event=command-received, command="{command}"')

        # create unique session id to be used as thread name
        # and stored as property of thread for logging emit
        session_id = self.create_session_id()

        # get the target method from command_map for the current command
        thread_method = self.command_map.get(command, self.process_invalid_command)
        self.service_thread(
            # use session_id as thread name to provide easy debugging per thread
            name=session_id,
            target=thread_method,
            args=(m,),
            session_id=session_id,
            trigger_id=trigger_id,
        )

    def process_heartbeat_command(self, message: dict) -> None:  # pylint: disable=unused-argument
        """Process the HeartBeat command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "command": "Heartbeat",
                "metric": {},
                "memoryPercent": 0,
                "cpuPercent": 0
            }

        Args:
            message: The message payload from the server topic.
        """
        self.heartbeat_watchdog = 0

        # send heartbeat -acknowledge- command
        response = {'command': 'Heartbeat', 'metric': self.metrics}
        self.message_broker.publish(
            message=json.dumps(response), topic=self.args.tc_svc_client_topic
        )
        self.log.info(f'feature=service, event=heartbeat-sent, metrics={self.metrics}')

    def process_logging_change_command(self, message: dict) -> None:
        """Process the LoggingChange command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "command": "LoggingChange",
                "level": "DEBUG"
            }

        Args:
            message: The message payload from the server topic.
        """
        level: str = message.get('level')
        self.log.info(f'feature=service, event=logging-change, level={level}')
        self.logger.update_handler_level(level)

    def process_invalid_command(self, message: dict) -> None:
        """Process all invalid commands.

        Args:
            message: The message payload from the server topic.
        """
        self.log.warning(
            f'feature=service, event=invalid-command-received, message="""({message})""".'
        )

    def process_shutdown_command(self, message: dict) -> None:
        """Implement parent method to process the shutdown command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "command": "Shutdown",
                "reason": "Service disabled by user."
            }

        Args:
            message: The message payload from the server topic.
        """
        reason = message.get('reason') or (
            'A shutdown command was received on server topic. Service is shutting down.'
        )
        self.log.info(f'feature=service, event=shutdown, reason={reason}')

        # acknowledge shutdown command
        self.message_broker.publish(
            json.dumps({'command': 'Acknowledged', 'type': 'Shutdown'}),
            self.args.tc_svc_client_topic,
        )

        # call App shutdown callback
        if callable(self.shutdown_callback):
            try:
                # call callback for shutdown and handle exceptions to protect thread
                self.shutdown_callback()  # pylint: disable=not-callable
            except Exception as e:
                self.log.error(
                    f'feature=service, event=shutdown-callback-error, error="""({e})""".'
                )
                self.log.trace(traceback.format_exc())

        # unsubscribe and disconnect from the broker
        self.message_broker.client.unsubscribe(self.args.tc_svc_server_topic)
        self.message_broker.client.disconnect()

        # update shutdown flag
        self.message_broker.shutdown = True

        # delay shutdown to give App time to cleanup
        time.sleep(5)
        self.tcex.exit(0)  # final shutdown in case App did not

    @property
    def ready(self) -> bool:
        """Return ready boolean."""
        return self._ready

    @ready.setter
    def ready(self, bool_val: bool):
        """Set ready boolean."""
        if isinstance(bool_val, bool) and bool_val is True:
            # wait until connected to send ready command
            while not self.message_broker._connected:
                if self.message_broker.shutdown:
                    break
                time.sleep(1)
            else:  # pylint: disable=useless-else-on-loop
                self.log.info('feature=service, event=service-ready')
                ready_command = {'command': 'Ready'}
                if self.ij.runtime_level.lower() in ['apiservice']:
                    ready_command['discoveryTypes'] = self.ij.service_discovery_types
                self.message_broker.publish(
                    json.dumps(ready_command), self.args.tc_svc_client_topic
                )
                self._ready = True

    @property
    def redis_client(self) -> object:
        """Return the correct KV store for this execution."""
        if self._redis_client is None:
            self._redis_client: object = self.tcex.redis_client
        return self._redis_client

    def service_thread(
        self,
        name: str,
        target: Callable[[], bool],
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
        session_id: Optional[str] = None,
        trigger_id: Optional[int] = None,
    ) -> None:
        """Start a message thread.

        Args:
            name: The name of the thread.
            target: The method to call for the thread.
            args: The args to pass to the target method.
            kwargs: Additional args.
            session_id: The current session id.
            trigger_id: The current trigger id.
        """
        self.log.info(f'feature=service, event=service-thread-creation, name={name}')
        args = args or ()
        try:
            t = threading.Thread(name=name, target=target, args=args, kwargs=kwargs, daemon=True)
            # add session_id to thread to use in logger emit
            t.session_id = session_id
            # add trigger_id to thread to use in logger emit
            t.trigger_id = trigger_id
            t.start()
        except Exception:
            self.log.trace(traceback.format_exc())

    @property
    def session_id(self) -> Optional[str]:
        """Return the current session_id."""
        if not hasattr(threading.current_thread(), 'session_id'):
            threading.current_thread().session_id = self.create_session_id()
        return threading.current_thread().session_id

    @property
    def thread_name(self) -> str:
        """Return a uuid4 session id."""
        return threading.current_thread().name

    @property
    def trigger_id(self) -> Optional[int]:
        """Return the current trigger_id."""
        trigger_id = None
        if hasattr(threading.current_thread(), 'trigger_id'):
            trigger_id = threading.current_thread().trigger_id
            if trigger_id is not None:
                trigger_id = int(trigger_id)
        return trigger_id

    def update_metric(self, label: str, value: Union[int, str]) -> None:
        """Update a metric if already exists.

        Args:
            label: The metric label (e.g., hits) to update.
            value: The updated value for the metric.
        """
        if self._metrics.get(label) is not None:
            self._metrics[label] = value



#!/usr/bin/env python
"""TcEx Framework Package Module."""
# standard library
import json
import os
import re
import shutil
import zipfile

# third-party
import colorama as c

from ..app_config_object import InstallJson
from .bin import Bin


class Package(Bin):
    """Package ThreatConnect Job or Playbook App for deployment.

    This method will package the app for deployment to ThreatConnect. Validation of the
    install.json file or files will be automatically run before packaging the app.

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Initialize Class properties."""
        super().__init__(_args)

        # properties
        self.features = ['aotExecutionEnabled', 'secureParams']

        # properties
        self._app_packages = []
        self.package_data = {'errors': [], 'updates': [], 'bundle': [], 'package': []}
        self.validation_data = {}

    @property
    def _build_excludes(self):
        """Return a list of files and folders that should be excluded during the build process."""
        return [
            'tcex.json',
            self.args.outdir,
            '__pycache__',
            '.cache',  # local cache directory
            '.c9',  # C9 IDE
            '.coverage',  # coverage file
            '.coveragerc',  # coverage configuration file file
            '.git',  # git directory
            '.gitmodules',  # git modules
            '.idea',  # PyCharm
            '.pytest_cache',  # pytest cache directory
            '*.iml',  # PyCharm files
            '*.pyc',  # any pyc file
            '.python-version',  # pyenv
            '.vscode',  # Visual Studio Code
            'artifacts',  # pytest in BB Pipelines
            'assets',  # pytest in BB Pipelines
            'local-*',  # log directory
            'log',  # log directory
            'test-reports',  # pytest in BB Pipelines
            'tests',  # pytest test directory
        ]

    def bundle(self, bundle_name):
        """Bundle multiple Job or Playbook Apps into a single zip file.

        Args:
            bundle_name (str): The output name of the bundle zip file.
        """
        if self.args.bundle or self.tj.package_bundle:
            if self.tj.package_bundle_packages:
                for bundle in self.tj.package_bundle_packages:
                    bundle_name = bundle.get('name')
                    bundle_patterns = bundle.get('patterns')

                    bundle_apps = []
                    for app in self._app_packages:
                        for app_pattern in bundle_patterns:
                            p = re.compile(app_pattern, re.IGNORECASE)
                            if p.match(app):
                                bundle_apps.append(app)

                    # bundle app in zip
                    if bundle_apps:
                        self.bundle_apps(bundle_name, bundle_apps)
            else:
                self.bundle_apps(bundle_name, self._app_packages)

    def bundle_apps(self, bundle_name, bundle_apps):
        """Bundle multiple Job or Playbook Apps (.tcx files) into a single zip file.

        Args:
            bundle_name (str): The output name of the bundle zip file.
            bundle_apps (list): A list of Apps to include in the bundle.
        """
        bundle_file = os.path.join(self.app_path, self.args.outdir, f'{bundle_name}-bundle.zip')
        z = zipfile.ZipFile(bundle_file, 'w')
        for app in bundle_apps:
            # update package data
            self.package_data['bundle'].append(
                {'action': 'Adding App:', 'output': os.path.basename(app)}
            )
            z.write(app, os.path.basename(app))

        # update package data
        self.package_data['bundle'].append(
            {'action': 'Created Bundle:', 'output': os.path.basename(bundle_file)}
        )
        z.close()

    def package(self):
        """Build the App package for deployment to ThreatConnect Exchange."""
        # create build directory
        tmp_path = os.path.join(self.app_path, self.args.outdir, 'build')
        if not os.path.isdir(tmp_path):
            os.makedirs(tmp_path)

        # temp path and cleanup
        template_app_path = os.path.join(tmp_path, 'template')
        if os.access(template_app_path, os.W_OK):
            # cleanup any previous failed builds
            shutil.rmtree(template_app_path)

        # update package data
        self.package_data['package'].append(
            {'action': 'Template Directory:', 'output': template_app_path}
        )

        # build exclude file/directory list
        excludes = list(self._build_excludes)
        excludes.extend(self.args.exclude)
        excludes.extend(self.tj.package_excludes)

        # update package data
        self.package_data['package'].append({'action': 'Excluded Files:', 'output': excludes})

        # copy project directory to temp location to use as template for multiple builds
        ignore_patterns = shutil.ignore_patterns(*excludes)
        shutil.copytree(self.app_path, template_app_path, False, ignore_patterns)

        # build list of app json files
        contents = os.listdir(self.app_path)
        if self.args.install_json is not None:
            contents = [self.args.install_json]

        # package app
        for install_json_name in sorted(contents):
            # skip files that are not install.json files
            if not install_json_name.endswith('install.json'):
                continue

            # load install json
            ij = InstallJson(install_json_name)

            # get App Name from config, install.json prefix or directory name.
            app_name = self.app_name(install_json_name)

            # update package data
            self.package_data['package'].append({'action': 'App Name:', 'output': app_name})

            # find a usable app version
            app_version = self.app_version(ij)

            # update package data
            self.package_data['package'].append(
                {'action': 'App Version:', 'output': f'{app_version}'}
            )

            # !!! The name of the folder in the zip is the *key* for an App. This value must
            # !!! remain consistent for the App to upgrade successfully.
            app_name_version = f'{app_name}_{app_version}'

            # build app directory
            tmp_app_path = os.path.join(tmp_path, app_name_version)
            if os.access(tmp_app_path, os.W_OK):
                # cleanup any previous failed builds
                shutil.rmtree(tmp_app_path)
            shutil.copytree(template_app_path, tmp_app_path)

            # Copy install.json
            # TODO: do we need copy if writing the data in the next step?
            shutil.copy(install_json_name, os.path.join(tmp_app_path, 'install.json'))

            # load template install json
            ij_template = InstallJson(path=tmp_app_path)

            # automatically update install.json in template directory
            ij_template.update(
                commit_hash=True, sequence=False, valid_values=False, playbook_data_types=False
            )

            # zip file
            self.zip_file(self.app_path, app_name_version, tmp_path)

            # cleanup build directory
            shutil.rmtree(tmp_app_path)

        # bundle zips (must have more than 1 app)
        if len(self._app_packages) > 1:
            self.bundle(self.tj.package_bundle_name or app_name)

    def app_name(self, install_json_name):
        """Return the app package name without version.

        1. Use the prefix on the install.json file (bundled Apps).
        2. Use the app_name field from the tcex.json file.
        3. Use the app directory name. This option should not be used.
        """
        if install_json_name != 'install.json':
            return install_json_name.split('.')[0]
        return self.tj.package_app_name or os.path.basename(self.app_path)

    def app_version(self, ij):
        """Return the app version "v1".

        1. Use app_version value from tcex.json if available. Typicall version is major only
           (e.g., v1), but from older Apps it could have minor version.
        2. Use major version from programVersion field in install.json if available.
        3. Default to '1.0.0' updated to major version only ('v1').
        """
        app_version = f"v{ij.program_version.split('.')[0]}"
        return self.tj.package_app_version or app_version

    def print_json(self):
        """Print JSON output containing results of the package command."""
        print(
            json.dumps({'package_data': self.package_data, 'validation_data': self.validation_data})
        )

    def print_results(self):
        """Print results of the package command."""
        # Updates
        if self.package_data.get('updates'):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Updates:')
            for p in self.package_data['updates']:
                print(
                    f"{p.get('action')!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {p.get('output')!s:<50}"
                )

        # Packaging
        print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Package:')
        for p in self.package_data['package']:
            if isinstance(p.get('output'), list):
                n = 5
                list_data = p.get('output')
                print(
                    f"{p.get('action'):<20}{c.Style.BRIGHT}{c.Fore.CYAN} "
                    f"{', '.join(p.get('output')[:n]):<50}"
                )
                del list_data[:n]
                for data in [
                    list_data[i : i + n] for i in range(0, len(list_data), n)  # noqa: E203
                ]:
                    print(f"{''!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {', '.join(data)!s:<50}")

            else:
                print(
                    f"{p.get('action')!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {p.get('output')!s:<50}"
                )

        # Bundle
        if self.package_data.get('bundle', False):
            print(f'\n{c.Style.BRIGHT}{c.Fore.BLUE}Bundle:')
            for p in self.package_data['bundle']:
                print(
                    f"{p.get('action')!s:<20}{c.Style.BRIGHT}{c.Fore.CYAN} {p.get('output')!s:<50}"
                )

        # ignore exit code
        if not self.args.ignore_validation:
            print('\n')  # separate errors from normal output
            # print all errors
            for error in self.package_data.get('errors'):
                print(f'{c.Fore.RED}{error}')
                self.exit_code = 1

    def zip_file(self, app_path, app_name, tmp_path):
        """Zip the App with tcex extension.

        Args:
            app_path (str): The path of the current project.
            app_name (str): The name of the App.
            tmp_path (str): The temp output path for the zip.
        """
        # zip build directory
        zip_file = os.path.join(app_path, self.args.outdir, app_name)
        zip_file_zip = f'{zip_file}.zip'
        zip_file_tcx = f'{zip_file}.tcx'
        shutil.make_archive(zip_file, 'zip', tmp_path, app_name)
        shutil.move(zip_file_zip, zip_file_tcx)
        self._app_packages.append(zip_file_tcx)
        # update package data
        self.package_data['package'].append({'action': 'App Package:', 'output': zip_file_tcx})



"""ThreatConnect Tag"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class Tags(CommonCaseManagementCollection):
    """Tags Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Tag. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Tag objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.TAGS,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

        if initial_response:
            for item in initial_response.get('data', []):
                if isinstance(item, dict):
                    item = Tag(tcex, **item)
                self.added_items.append(item)

    def __iter__(self):
        """Iterate on Tag Collection"""
        return self.iterate(initial_response=self.initial_response)

    def add_tag(self, tag):
        """Add a Tag.

        Args:
            tag (Tag): The Data Object to add.
        """
        self.added_items.append(tag)

    def entity_map(self, entity):
        """Map a dict to a Tag.

        Args:
            entity (dict): The Tag data.

        Returns:
            CaseManagement.Tag: An Tag Object
        """
        return Tag(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterTags Object."""
        return FilterTags(ApiEndpoints.TAGS, self.tcex, self.tql)


class Tag(CommonCaseManagement):
    """Tag object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        cases (Case, kwargs): [Read-Only] The **Cases** for the Tag.
        description (str, kwargs): a brief description of the Tag
        last_used (str, kwargs): [Read-Only] The **Last Used** for the Tag.
        name (str, kwargs): [Required] The **Name** for the Tag.
        owner (str, kwargs): [Read-Only] The **Owner** for the Tag.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.TAGS, kwargs)
        self._cases = kwargs.get('cases', None)
        self._description = kwargs.get('description', None)
        self._last_used = kwargs.get('last_used', None)
        self._name = kwargs.get('name', None)
        self._owner = kwargs.get('owner', None)

    @property
    def as_entity(self):
        """Return the entity representation of the Tag."""
        return {'type': 'Tag', 'value': self.name, 'id': self.id}

    @property
    def cases(self):
        """Return the **Cases** for the Tag."""
        if self._cases:
            return self.tcex.cm.cases(initial_response=self._cases)
        return self._cases

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_case = Tag(self.tcex, **entity)
        self.__dict__.update(new_case.__dict__)

    @property
    def description(self):
        """Return the **Description** for the Tag."""
        return self._description

    @description.setter
    def description(self, description):
        """Set the **Description** for the Tag."""
        self._description = description

    @property
    def last_used(self):
        """Return the **Last Used** for the Tag."""
        return self._last_used

    @property
    def name(self):
        """Return the **Name** for the Tag."""
        return self._name

    @name.setter
    def name(self, name):
        """Set the **Name** for the Tag."""
        self._name = name

    @property
    def owner(self):
        """Return the **Owner** for the Tag."""
        return self._owner


class FilterTags(Filter):
    """Filter Object for Tags"""

    def case_id(self, operator, case_id):
        """Filter Tags based on **caseId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_id (int): The ID of the case the tag is applied to.
        """
        self._tql.add_filter('caseId', operator, case_id, TQL.Type.INTEGER)

    def description(self, operator, description):
        """Filter Tags based on **description** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            description (str): The description of the tag.
        """
        self._tql.add_filter('description', operator, description, TQL.Type.STRING)

    @property
    def has_case(self):
        """Return **FilterCases** for further filtering."""
        from .case import FilterCases

        cases = FilterCases(ApiEndpoints.CASES, self._tcex, TQL())
        self._tql.add_filter('hasCase', TQL.Operator.EQ, cases, TQL.Type.SUB_QUERY)
        return cases

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Tags based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the tag.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def last_used(self, operator, last_used):
        """Filter Tags based on **lastUsed** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            last_used (str): The date this tag was last used.
        """
        self._tql.add_filter('lastUsed', operator, last_used, TQL.Type.STRING)

    def name(self, operator, name):
        """Filter Tags based on **name** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            name (str): The name of the tag.
        """
        self._tql.add_filter('name', operator, name, TQL.Type.STRING)

    def owner(self, operator, owner):
        """Filter Tags based on **owner** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            owner (int): The ID of the tag's Organization.
        """
        self._tql.add_filter('owner', operator, owner, TQL.Type.INTEGER)

    def owner_name(self, operator, owner_name):
        """Filter Tags based on **ownerName** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            owner_name (str): The name of the tag's Organization.
        """
        self._tql.add_filter('ownerName', operator, owner_name, TQL.Type.STRING)
